@@ -5,6 +5,50 @@ from typing import Any
 
 from .fpyast import *
 
+_expr_dispatch: dict[type[Expr], str] = {
+    Var: "_visit_var",
+    BoolVal: "_visit_bool",
+    ForeignVal: "_visit_foreign",
+    Decnum: "_visit_decnum",
+    Hexnum: "_visit_hexnum",
+    Integer: "_visit_integer",
+    Rational: "_visit_rational",
+    Digits: "_visit_digits",
+    NullaryOp: "_visit_nullaryop",
+    UnaryOp: "_visit_unaryop",
+    BinaryOp: "_visit_binaryop",
+    TernaryOp: "_visit_ternaryop",
+    NaryOp: "_visit_naryop",
+    Compare: "_visit_compare",
+    Call: "_visit_call",
+    TupleExpr: "_visit_tuple_expr",
+    ListExpr: "_visit_list_expr",
+    ListComp: "_visit_list_comp",
+    ListRef: "_visit_list_ref",
+    ListSlice: "_visit_list_slice",
+    ListSet: "_visit_list_set",
+    IfExpr: "_visit_if_expr",
+    Attribute: "_visit_attribute",
+
+    Round: "_visit_round",
+    RoundExact: "_visit_round",
+    RoundAt: "_visit_round_at",
+}
+
+_stmt_dispatch: dict[type[Stmt], str] = {
+    Assign: "_visit_assign",
+    IndexedAssign: "_visit_indexed_assign",
+    If1Stmt: "_visit_if1",
+    IfStmt: "_visit_if",
+    WhileStmt: "_visit_while",
+    ForStmt: "_visit_for",
+    ContextStmt: "_visit_context",
+    AssertStmt: "_visit_assert",
+    EffectStmt: "_visit_effect",
+    ReturnStmt: "_visit_return",
+    PassStmt: "_visit_pass",
+}
+
 class Visitor(ABC):
     """
     Visitor base class for FPy AST nodes.
@@ -46,7 +90,7 @@ class Visitor(ABC):
         ...
 
     @abstractmethod
-    def _visit_constant(self, e: Constant, ctx: Any) -> Any:
+    def _visit_nullaryop(self, e: NullaryOp, ctx: Any) -> Any:
         ...
 
     @abstractmethod
@@ -65,28 +109,42 @@ class Visitor(ABC):
     def _visit_naryop(self, e: NaryOp, ctx: Any) -> Any:
         ...
 
-    @abstractmethod
-    def _visit_compare(self, e: Compare, ctx: Any) -> Any:
-        ...
+    def _visit_round(self, e: Round | RoundExact, ctx: Any) -> Any:
+        return self._visit_unaryop(e, ctx)
+
+    def _visit_round_at(self, e: RoundAt, ctx: Any) -> Any:
+        return self._visit_binaryop(e, ctx)
 
     @abstractmethod
     def _visit_call(self, e: Call, ctx: Any) -> Any:
         ...
 
+    @abstractmethod
+    def _visit_compare(self, e: Compare, ctx: Any) -> Any:
+        ...
+
     @abstractmethod
     def _visit_tuple_expr(self, e: TupleExpr, ctx: Any) -> Any:
         ...
 
     @abstractmethod
-    def _visit_comp_expr(self, e: CompExpr, ctx: Any) -> Any:
+    def _visit_list_expr(self, e: ListExpr, ctx: Any) -> Any:
         ...
 
     @abstractmethod
-    def _visit_tuple_ref(self, e: TupleRef, ctx: Any) -> Any:
+    def _visit_list_comp(self, e: ListComp, ctx: Any) -> Any:
         ...
 
     @abstractmethod
-    def _visit_tuple_set(self, e: TupleSet, ctx: Any) -> Any:
+    def _visit_list_ref(self, e: ListRef, ctx: Any) -> Any:
+        ...
+
+    @abstractmethod
+    def _visit_list_slice(self, e: ListSlice, ctx: Any) -> Any:
+        ...
+
+    @abstractmethod
+    def _visit_list_set(self, e: ListSet, ctx: Any) -> Any:
         ...
 
     @abstractmethod
@@ -94,7 +152,7 @@ class Visitor(ABC):
         ...
 
     @abstractmethod
-    def _visit_context_expr(self, e: ContextExpr, ctx: Any) -> Any:
+    def _visit_attribute(self, e: Attribute, ctx: Any) -> Any:
         ...
 
     #######################################################
@@ -140,6 +198,9 @@ class Visitor(ABC):
     def _visit_return(self, stmt: ReturnStmt, ctx: Any) -> Any:
         ...
 
+    @abstractmethod
+    def _visit_pass(self, stmt: PassStmt, ctx: Any) -> Any:
+        ...
 
     #######################################################
     # Block
@@ -160,77 +221,26 @@ class Visitor(ABC):
 
     def _visit_expr(self, e: Expr, ctx: Any) -> Any:
         """Dispatch to the appropriate visit method for an expression."""
-        match e:
-            case Var():
-                return self._visit_var(e, ctx)
-            case BoolVal():
-                return self._visit_bool(e, ctx)
-            case ForeignVal():
-                return self._visit_foreign(e, ctx)
-            case Decnum():
-                return self._visit_decnum(e, ctx)
-            case Hexnum():
-                return self._visit_hexnum(e, ctx)
-            case Integer():
-                return self._visit_integer(e, ctx)
-            case Rational():
-                return self._visit_rational(e, ctx)
-            case Digits():
-                return self._visit_digits(e, ctx)
-            case Constant():
-                return self._visit_constant(e, ctx)
-            case UnaryOp():
-                return self._visit_unaryop(e, ctx)
-            case BinaryOp():
-                return self._visit_binaryop(e, ctx)
-            case TernaryOp():
-                return self._visit_ternaryop(e, ctx)
-            case NaryOp():
-                return self._visit_naryop(e, ctx)
-            case Compare():
-                return self._visit_compare(e, ctx)
-            case Call():
-                return self._visit_call(e, ctx)
-            case TupleExpr():
-                return self._visit_tuple_expr(e, ctx)
-            case CompExpr():
-                return self._visit_comp_expr(e, ctx)
-            case TupleRef():
-                return self._visit_tuple_ref(e, ctx)
-            case TupleSet():
-                return self._visit_tuple_set(e, ctx)
-            case IfExpr():
-                return self._visit_if_expr(e, ctx)
-            case ContextExpr():
-                return self._visit_context_expr(e, ctx)
-            case _:
-                raise NotImplementedError(f'unreachable {e}')
+        # Walk the inheritance chain until we find a dispatch method or reach Expr
+        for cls in type(e).__mro__:
+            if cls is Expr:
+                break
+            method = _expr_dispatch.get(cls, None)
+            if method is not None:
+                func = getattr(self, method)
+                return func(e, ctx)
+        raise NotImplementedError(f'no dispatch method for `{e}`')
 
     def _visit_statement(self, stmt: Stmt, ctx: Any) -> Any:
         """Dispatch to the appropriate visit method for a statement."""
-        match stmt:
-            case Assign():
-                return self._visit_assign(stmt, ctx)
-            case IndexedAssign():
-                return self._visit_indexed_assign(stmt, ctx)
-            case If1Stmt():
-                return self._visit_if1(stmt, ctx)
-            case IfStmt():
-                return self._visit_if(stmt, ctx)
-            case WhileStmt():
-                return self._visit_while(stmt, ctx)
-            case ForStmt():
-                return self._visit_for(stmt, ctx)
-            case ContextStmt():
-                return self._visit_context(stmt, ctx)
-            case AssertStmt():
-                return self._visit_assert(stmt, ctx)
-            case EffectStmt():
-                return self._visit_effect(stmt, ctx)
-            case ReturnStmt():
-                return self._visit_return(stmt, ctx)
-            case _:
-                raise NotImplementedError(f'unreachable: {stmt}')
+        for cls in type(stmt).__mro__:
+            if cls is Expr:
+                break
+            method = _stmt_dispatch.get(cls, None)
+            if method is not None:
+                func = getattr(self, method)
+                return func(stmt, ctx)
+        raise NotImplementedError(f'no dispatch method for `{stmt}`')
 
 #####################################################################
 # Default visitor
@@ -259,10 +269,10 @@ class DefaultVisitor(Visitor):
     def _visit_rational(self, e: Rational, ctx: Any):
         pass
 
-    def _visit_constant(self, e: Constant, ctx: Any):
+    def _visit_digits(self, e: Digits, ctx: Any):
         pass
 
-    def _visit_digits(self, e: Digits, ctx: Any):
+    def _visit_nullaryop(self, e: NullaryOp, ctx: Any):
         pass
 
     def _visit_unaryop(self, e: UnaryOp, ctx: Any):
@@ -281,30 +291,40 @@ class DefaultVisitor(Visitor):
         for arg in e.args:
             self._visit_expr(arg, ctx)
 
-    def _visit_compare(self, e: Compare, ctx: Any):
-        for c in e.args:
-            self._visit_expr(c, ctx)
-
     def _visit_call(self, e: Call, ctx: Any):
         for arg in e.args:
             self._visit_expr(arg, ctx)
 
-    def _visit_tuple_expr(self, e: TupleExpr, ctx: Any):
+    def _visit_compare(self, e: Compare, ctx: Any):
         for c in e.args:
             self._visit_expr(c, ctx)
 
-    def _visit_tuple_ref(self, e: TupleRef, ctx: Any):
-        self._visit_expr(e.value, ctx)
-        for s in e.slices:
-            self._visit_expr(s, ctx)
+    def _visit_tuple_expr(self, e: TupleExpr, ctx: Any):
+        for c in e.elts:
+            self._visit_expr(c, ctx)
 
-    def _visit_tuple_set(self, e: TupleSet, ctx: Any):
-        self._visit_expr(e.array, ctx)
-        for s in e.slices:
-            self._visit_expr(s, ctx)
-        self._visit_expr(e.value, ctx)
+    def _visit_list_expr(self, e: ListExpr, ctx: Any):
+        for c in e.elts:
+            self._visit_expr(c, ctx)
 
-    def _visit_comp_expr(self, e: CompExpr, ctx: Any):
+    def _visit_list_ref(self, e: ListRef, ctx: Any):
+        self._visit_expr(e.value, ctx)
+        self._visit_expr(e.index, ctx)
+
+    def _visit_list_slice(self, e: ListSlice, ctx: Any):
+        self._visit_expr(e.value, ctx)
+        if e.start is not None:
+            self._visit_expr(e.start, ctx)
+        if e.stop is not None:
+            self._visit_expr(e.stop, ctx)
+
+    def _visit_list_set(self, e: ListSet, ctx: Any):
+        self._visit_expr(e.value, ctx)
+        for s in e.indices:
+            self._visit_expr(s, ctx)
+        self._visit_expr(e.expr, ctx)
+
+    def _visit_list_comp(self, e: ListComp, ctx: Any):
         for iterable in e.iterables:
             self._visit_expr(iterable, ctx)
         self._visit_expr(e.elt, ctx)
@@ -314,16 +334,14 @@ class DefaultVisitor(Visitor):
         self._visit_expr(e.ift, ctx)
         self._visit_expr(e.iff, ctx)
 
-    def _visit_context_expr(self, e: ContextExpr, ctx: Any):
-        for arg in e.args:
-            if not isinstance(arg, ForeignAttribute):
-                self._visit_expr(arg, ctx)
+    def _visit_attribute(self, e: Attribute, ctx: Any):
+        self._visit_expr(e.value, ctx)
 
     def _visit_assign(self, stmt: Assign, ctx: Any):
         self._visit_expr(stmt.expr, ctx)
 
     def _visit_indexed_assign(self, stmt: IndexedAssign, ctx: Any):
-        for s in stmt.slices:
+        for s in stmt.indices:
             self._visit_expr(s, ctx)
         self._visit_expr(stmt.expr, ctx)
 
@@ -350,6 +368,8 @@ class DefaultVisitor(Visitor):
 
     def _visit_assert(self, stmt: AssertStmt, ctx: Any):
         self._visit_expr(stmt.test, ctx)
+        if stmt.msg is not None:
+            self._visit_expr(stmt.msg, ctx)
 
     def _visit_effect(self, stmt: EffectStmt, ctx: Any):
         self._visit_expr(stmt.expr, ctx)
@@ -357,6 +377,9 @@ class DefaultVisitor(Visitor):
     def _visit_return(self, stmt: ReturnStmt, ctx: Any):
         self._visit_expr(stmt.expr, ctx)
 
+    def _visit_pass(self, stmt: PassStmt, ctx: Any):
+        pass
+
     def _visit_block(self, block: StmtBlock, ctx: Any):
         for stmt in block.stmts:
             self._visit_statement(stmt, ctx)
@@ -383,67 +406,90 @@ class DefaultTransformVisitor(Visitor):
         return Decnum(e.val, e.loc)
 
     def _visit_hexnum(self, e: Hexnum, ctx: Any):
-        return Hexnum(e.val, e.loc)
+        return Hexnum(e.func, e.val, e.loc)
 
     def _visit_integer(self, e: Integer, ctx: Any):
         return Integer(e.val, e.loc)
 
     def _visit_rational(self, e: Rational, ctx: Any):
-        return Rational(e.p, e.q, e.loc)
-
-    def _visit_constant(self, e: Constant, ctx: Any):
-        return Constant(e.val, e.loc)
+        return Rational(e.func, e.p, e.q, e.loc)
 
     def _visit_digits(self, e: Digits, ctx: Any):
-        return Digits(e.m, e.e, e.b, e.loc)
+        return Digits(e.func, e.m, e.e, e.b, e.loc)
+
+    def _visit_nullaryop(self, e: NullaryOp, ctx: Any):
+        return type(e)(e.func, e.loc)
 
     def _visit_unaryop(self, e: UnaryOp, ctx: Any):
         arg = self._visit_expr(e.arg, ctx)
-        return type(e)(arg, e.loc)
+        if isinstance(e, NamedUnaryOp):
+            return type(e)(e.func, arg, e.loc)
+        else:
+            return type(e)(arg, e.loc)
 
     def _visit_binaryop(self, e: BinaryOp, ctx: Any):
         first = self._visit_expr(e.first, ctx)
         second = self._visit_expr(e.second, ctx)
-        return type(e)(first, second, e.loc)
+        if isinstance(e, NamedBinaryOp):
+            return type(e)(e.func, first, second, e.loc)
+        else:
+            return type(e)(first, second, e.loc)
 
     def _visit_ternaryop(self, e: TernaryOp, ctx: Any):
         first = self._visit_expr(e.first, ctx)
         second = self._visit_expr(e.second, ctx)
         third = self._visit_expr(e.third, ctx)
-        return type(e)(first, second, third, e.loc)
+        if isinstance(e, NamedTernaryOp):
+            return type(e)(e.func, first, second, third, e.loc)
+        else:
+            return type(e)(first, second, third, e.loc)
 
     def _visit_naryop(self, e: NaryOp, ctx: Any):
         args = [self._visit_expr(arg, ctx) for arg in e.args]
-        return type(e)(args, e.loc)
+        if isinstance(e, NamedNaryOp):
+            return type(e)(e.func, args, e.loc)
+        else:
+            return type(e)(args, e.loc)
+
+    def _visit_call(self, e: Call, ctx: Any):
+        args = [self._visit_expr(arg, ctx) for arg in e.args]
+        kwargs = [ (k, self._visit_expr(v, ctx)) for k, v in e.kwargs ]
+        return Call(e.func, e.fn, args, kwargs, e.loc)
 
     def _visit_compare(self, e: Compare, ctx: Any):
         args = [self._visit_expr(arg, ctx) for arg in e.args]
         return Compare(e.ops, args, e.loc)
 
-    def _visit_call(self, e: Call, ctx: None):
-        args = [self._visit_expr(arg, ctx) for arg in e.args]
-        return Call(e.name, args, e.loc)
-
     def _visit_tuple_expr(self, e: TupleExpr, ctx: Any):
-        args = [self._visit_expr(arg, ctx) for arg in e.args]
+        args = [self._visit_expr(arg, ctx) for arg in e.elts]
         return TupleExpr(args, e.loc)
 
-    def _visit_tuple_ref(self, e: TupleRef, ctx: Any):
-        value = self._visit_expr(e.value, ctx)
-        slices = [self._visit_expr(s, ctx) for s in e.slices]
-        return TupleRef(value, slices, e.loc)
+    def _visit_list_expr(self, e: ListExpr, ctx: Any):
+        args = [self._visit_expr(arg, ctx) for arg in e.elts]
+        return ListExpr(args, e.loc)
 
-    def _visit_tuple_set(self, e: TupleSet, ctx: Any):
-        array = self._visit_expr(e.array, ctx)
-        slices = [self._visit_expr(s, ctx) for s in e.slices]
+    def _visit_list_ref(self, e: ListRef, ctx: Any):
         value = self._visit_expr(e.value, ctx)
-        return TupleSet(array, slices, value, e.loc)
+        index = self._visit_expr(e.index, ctx)
+        return ListRef(value, index, e.loc)
 
-    def _visit_comp_expr(self, e: CompExpr, ctx: Any):
+    def _visit_list_slice(self, e: ListSlice, ctx: Any):
+        value = self._visit_expr(e.value, ctx)
+        start = None if e.start is None else self._visit_expr(e.start, ctx)
+        stop = None if e.stop is None else self._visit_expr(e.stop, ctx)
+        return ListSlice(value, start, stop, e.loc)
+
+    def _visit_list_set(self, e: ListSet, ctx: Any):
+        array = self._visit_expr(e.value, ctx)
+        slices = [self._visit_expr(s, ctx) for s in e.indices]
+        value = self._visit_expr(e.expr, ctx)
+        return ListSet(array, slices, value, e.loc)
+
+    def _visit_list_comp(self, e: ListComp, ctx: Any):
         targets = [self._visit_binding(target, ctx) for target in e.targets]
         iterables = [self._visit_expr(iterable, ctx) for iterable in e.iterables]
         elt = self._visit_expr(e.elt, ctx)
-        return CompExpr(targets, iterables, elt, e.loc)
+        return ListComp(targets, iterables, elt, e.loc)
 
     def _visit_if_expr(self, e: IfExpr, ctx: Any):
         cond = self._visit_expr(e.cond, ctx)
@@ -451,32 +497,9 @@ class DefaultTransformVisitor(Visitor):
         iff = self._visit_expr(e.iff, ctx)
         return IfExpr(cond, ift, iff, e.loc)
 
-    def _visit_context_expr(self, e: ContextExpr, ctx: Any):
-        match e.ctor:
-            case Var():
-                ctor = self._visit_var(e.ctor, ctx)
-            case ForeignAttribute():
-                ctor = ForeignAttribute(e.ctor.name, e.ctor.attrs, e.loc)
-            case _:
-                raise RuntimeError('unreachable', e.ctor)
-
-        args: list[Expr | ForeignAttribute] = []
-        for arg in e.args:
-            match arg:
-                case ForeignAttribute():
-                    args.append(ForeignAttribute(arg.name, arg.attrs, arg.loc))
-                case _:
-                    args.append(self._visit_expr(arg, ctx))
-
-        kwargs: list[tuple[str, Expr | ForeignAttribute]] = []
-        for name, arg in e.kwargs:
-            match arg:
-                case ForeignAttribute():
-                    kwargs.append((name, ForeignAttribute(arg.name, arg.attrs, arg.loc)))
-                case _:
-                    kwargs.append((name, self._visit_expr(arg, ctx)))
-
-        return ContextExpr(ctor, args, kwargs, e.loc)
+    def _visit_attribute(self, e: Attribute, ctx: Any):
+        value = self._visit_expr(e.value, ctx)
+        return Attribute(value, e.attr, ctx)
 
     def _visit_binding(self, binding: Id | TupleBinding, ctx: Any):
         match binding:
@@ -492,13 +515,13 @@ class DefaultTransformVisitor(Visitor):
         return TupleBinding(elts, binding.loc)
 
     def _visit_assign(self, stmt: Assign, ctx: Any):
-        binding = self._visit_binding(stmt.binding, ctx)
+        binding = self._visit_binding(stmt.target, ctx)
         expr = self._visit_expr(stmt.expr, ctx)
         s = Assign(binding, stmt.type, expr, stmt.loc)
         return s, ctx
 
     def _visit_indexed_assign(self, stmt: IndexedAssign, ctx: Any):
-        slices = [self._visit_expr(s, ctx) for s in stmt.slices]
+        slices = [self._visit_expr(s, ctx) for s in stmt.indices]
         expr = self._visit_expr(stmt.expr, ctx)
         s = IndexedAssign(stmt.var, slices, expr, stmt.loc)
         return s, ctx
@@ -530,24 +553,20 @@ class DefaultTransformVisitor(Visitor):
         return s, ctx
 
     def _visit_context(self, stmt: ContextStmt, ctx: Any):
-        match stmt.ctx:
-            case Var():
-                context = self._visit_var(stmt.ctx, ctx)
-            case ContextExpr():
-                context = self._visit_context_expr(stmt.ctx, ctx)
-            case ForeignVal():
-                context = ForeignVal(stmt.ctx.val, stmt.loc)
-            case _:
-                raise RuntimeError('unreachable', stmt.ctx)
-
+        context = self._visit_expr(stmt.ctx, ctx)
         body, _ = self._visit_block(stmt.body, ctx)
-        s = ContextStmt(stmt.name, context, body, stmt.loc)
+        s = ContextStmt(stmt.target, context, body, stmt.loc)
         return s, ctx
 
     def _visit_assert(self, stmt: AssertStmt, ctx: Any):
         test = self._visit_expr(stmt.test, ctx)
-        s = AssertStmt(test, stmt.msg, stmt.loc)
-        return s, ctx
+        if stmt.msg is None:
+            s = AssertStmt(test, None, stmt.loc)
+            return s, ctx
+        else:
+            msg = self._visit_expr(stmt.msg, ctx)
+            s = AssertStmt(test, msg, stmt.loc)
+            return s, ctx
 
     def _visit_effect(self, stmt: EffectStmt, ctx: Any):
         expr = self._visit_expr(stmt.expr, ctx)
@@ -559,6 +578,10 @@ class DefaultTransformVisitor(Visitor):
         s = ReturnStmt(expr, stmt.loc)
         return s, ctx
 
+    def _visit_pass(self, stmt: PassStmt, ctx: Any):
+        s = PassStmt(stmt.loc)
+        return s, ctx
+
     def _visit_block(self, block: StmtBlock, ctx: Any):
         stmts: list[Stmt] = []
         for stmt in block.stmts:
@@ -574,10 +597,12 @@ class DefaultTransformVisitor(Visitor):
         return FuncDef(
             func.name,
             args,
+            func.free_vars,
+            func.ctx,
             body,
-            metadata=func.metadata,
-            free_vars=func.free_vars,
-            ctx=func.ctx,
+            func.spec,
+            func.meta,
+            func.env,
             loc=func.loc
         )
 
