@@ -1,6 +1,4 @@
-"""Compilation from FPy IR to FPCore"""
-
-from typing import Optional
+"""Compilation from FPy to FPCore."""
 
 import titanfp.fpbench.fpcast as fpc
 
@@ -9,80 +7,125 @@ from ..ast import *
 from ..fpc_context import FPCoreContext
 from ..function import Function
 from ..number import Context
-from ..transform import ContextInline, ForBundling, ForUnpack, FuncUpdate, IfBundling, WhileBundling
+from ..transform import ConstFold, ForBundling, ForUnpack, FuncUpdate, IfBundling, WhileBundling
 from ..utils import Gensym
 
-from .backend import Backend
+from .backend import Backend, CompileError
 
-_unary_table: dict[type[UnaryOp], type[fpc.Expr]] = {
-    Fabs: fpc.Fabs,
-    Sqrt: fpc.Sqrt,
-    Neg: fpc.Neg,
-    Cbrt: fpc.Cbrt,
-    Ceil: fpc.Ceil,
-    Floor: fpc.Floor,
-    NearbyInt: fpc.Nearbyint,
-    Round: fpc.Round,
-    Trunc: fpc.Trunc,
-    Acos: fpc.Acos,
-    Asin: fpc.Asin,
-    Atan: fpc.Atan,
-    Cos: fpc.Cos,
-    Sin: fpc.Sin,
-    Tan: fpc.Tan,
-    Acosh: fpc.Acosh,
-    Asinh: fpc.Asinh,
-    Atanh: fpc.Atanh,
-    Cosh: fpc.Cosh,
-    Sinh: fpc.Sinh,
-    Tanh: fpc.Tanh,
-    Exp: fpc.Exp,
-    Exp2: fpc.Exp2,
-    Expm1: fpc.Expm1,
-    Log: fpc.Log,
-    Log10: fpc.Log10,
-    Log1p: fpc.Log1p,
-    Log2: fpc.Log2,
-    Erf: fpc.Erf,
-    Erfc: fpc.Erfc,
-    Lgamma: fpc.Lgamma,
-    Tgamma: fpc.Tgamma,
-    IsFinite: fpc.Isfinite,
-    IsInf: fpc.Isinf,
-    IsNan: fpc.Isnan,
-    IsNormal: fpc.Isnormal,
-    Signbit: fpc.Signbit,
-    Not: fpc.Not,
-    # rounding
-    Cast: fpc.Cast,
-}
+# Cached table storage
+_nullary_table_cache: dict[type[NullaryOp], fpc.Expr] | None = None
+_unary_table_cache: dict[type[UnaryOp], type[fpc.Expr]] | None = None
+_binary_table_cache: dict[type[BinaryOp], type[fpc.Expr]] | None = None
+_ternary_table_cache: dict[type[TernaryOp], type[fpc.Expr]] | None = None
+_nary_table_cache: dict[type[NaryOp], type[fpc.Expr]] | None = None
 
-_binary_table: dict[type[BinaryOp], type[fpc.Expr]] = {
-    Add: fpc.Add,
-    Sub: fpc.Sub,
-    Mul: fpc.Mul,
-    Div: fpc.Div,
-    Copysign: fpc.Copysign,
-    Fdim: fpc.Fdim,
-    Fmax: fpc.Fmax,
-    Fmin: fpc.Fmin,
-    Fmod: fpc.Fmod,
-    Remainder: fpc.Remainder,
-    Hypot: fpc.Hypot,
-    Atan2: fpc.Atan2,
-    Pow: fpc.Pow,
-}
+def _get_nullary_table() -> dict[type[NullaryOp], fpc.Expr]:
+    """Get the cached nullary operations table."""
+    global _nullary_table_cache
+    if _nullary_table_cache is None:
+        _nullary_table_cache = {
+            ConstNan: fpc.Constant('NAN'),
+            ConstInf: fpc.Constant('INFINITY'),
+            ConstPi: fpc.Constant('PI'),
+            ConstE: fpc.Constant('E'),
+            ConstLog2E: fpc.Constant('LOG2E'),
+            ConstLog10E: fpc.Constant('LOG10E'),
+            ConstLn2: fpc.Constant('LN2'),
+            ConstPi_2: fpc.Constant('PI_2'),
+            ConstPi_4: fpc.Constant('PI_4'),
+            Const1_Pi: fpc.Constant('M_1_PI'),
+            Const2_Pi: fpc.Constant('M_2_PI'),
+            Const2_SqrtPi: fpc.Constant('M_2_SQRTPI'),
+            ConstSqrt2: fpc.Constant('SQRT2'),
+            ConstSqrt1_2: fpc.Constant('SQRT1_2'),
+        }
+    return _nullary_table_cache
 
-_ternary_table: dict[type[TernaryOp], type[fpc.Expr]] = {
-    Fma: fpc.Fma,
-}
+def _get_unary_table() -> dict[type[UnaryOp], type[fpc.Expr]]:
+    """Get the cached unary operations table."""
+    global _unary_table_cache
+    if _unary_table_cache is None:
+        _unary_table_cache = {
+            Fabs: fpc.Fabs,
+            Sqrt: fpc.Sqrt,
+            Neg: fpc.Neg,
+            Cbrt: fpc.Cbrt,
+            Ceil: fpc.Ceil,
+            Floor: fpc.Floor,
+            NearbyInt: fpc.Nearbyint,
+            RoundInt: fpc.Round,
+            Trunc: fpc.Trunc,
+            Acos: fpc.Acos,
+            Asin: fpc.Asin,
+            Atan: fpc.Atan,
+            Cos: fpc.Cos,
+            Sin: fpc.Sin,
+            Tan: fpc.Tan,
+            Acosh: fpc.Acosh,
+            Asinh: fpc.Asinh,
+            Atanh: fpc.Atanh,
+            Cosh: fpc.Cosh,
+            Sinh: fpc.Sinh,
+            Tanh: fpc.Tanh,
+            Exp: fpc.Exp,
+            Exp2: fpc.Exp2,
+            Expm1: fpc.Expm1,
+            Log: fpc.Log,
+            Log10: fpc.Log10,
+            Log1p: fpc.Log1p,
+            Log2: fpc.Log2,
+            Erf: fpc.Erf,
+            Erfc: fpc.Erfc,
+            Lgamma: fpc.Lgamma,
+            Tgamma: fpc.Tgamma,
+            IsFinite: fpc.Isfinite,
+            IsInf: fpc.Isinf,
+            IsNan: fpc.Isnan,
+            IsNormal: fpc.Isnormal,
+            Signbit: fpc.Signbit,
+            Not: fpc.Not,
+        }
+    return _unary_table_cache
 
-_nary_table: dict[type[NaryOp], type[fpc.Expr]] = {
-    Or: fpc.Or,
-    And: fpc.And,
-}
+def _get_binary_table() -> dict[type[BinaryOp], type[fpc.Expr]]:
+    """Get the cached binary operations table."""
+    global _binary_table_cache
+    if _binary_table_cache is None:
+        _binary_table_cache = {
+            Add: fpc.Add,
+            Sub: fpc.Sub,
+            Mul: fpc.Mul,
+            Div: fpc.Div,
+            Copysign: fpc.Copysign,
+            Fdim: fpc.Fdim,
+            Fmod: fpc.Fmod,
+            Remainder: fpc.Remainder,
+            Hypot: fpc.Hypot,
+            Atan2: fpc.Atan2,
+            Pow: fpc.Pow,
+        }
+    return _binary_table_cache
 
-class FPCoreCompileError(Exception):
+def _get_ternary_table() -> dict[type[TernaryOp], type[fpc.Expr]]:
+    """Get the cached ternary operations table."""
+    global _ternary_table_cache
+    if _ternary_table_cache is None:
+        _ternary_table_cache = {
+            Fma: fpc.Fma,
+        }
+    return _ternary_table_cache
+
+def _get_nary_table() -> dict[type[NaryOp], type[fpc.Expr]]:
+    """Get the cached n-ary operations table."""
+    global _nary_table_cache
+    if _nary_table_cache is None:
+        _nary_table_cache = {
+            Or: fpc.Or,
+            And: fpc.And,
+        }
+    return _nary_table_cache
+
+class FPCoreCompileError(CompileError):
     """Any FPCore compilation error"""
     pass
 
@@ -100,16 +143,20 @@ def _size0_expr(x: str):
     return fpc.Size(fpc.Var(x), fpc.Integer(0))
 
 
-class FPCoreCompileInstance(Visitor):
+class _FPCoreCompileInstance(Visitor):
     """Compilation instance from FPy to FPCore"""
+
     func: FuncDef
     def_use: DefineUseAnalysis
     gensym: Gensym
 
-    def __init__(self, func: FuncDef, def_use: DefineUseAnalysis):
+    unsafe_int_cast: bool
+
+    def __init__(self, func: FuncDef, def_use: DefineUseAnalysis, unsafe_int_cast: bool = True):
         self.func = func
         self.def_use = def_use
-        self.gensym = Gensym(reserved=set(def_use.defs.keys()))
+        self.gensym = Gensym(reserved=def_use.names())
+        self.unsafe_int_cast = unsafe_int_cast
 
     def compile(self) -> fpc.FPCore:
         f = self._visit_function(self.func, None)
@@ -177,26 +224,83 @@ class FPCoreCompileInstance(Visitor):
         raise FPCoreCompileError('unsupported value', e.val)
 
     def _visit_decnum(self, e: Decnum, ctx: None) -> fpc.Expr:
-        return fpc.Decnum(e.val)
+        if e.is_integer() and self.unsafe_int_cast:
+            # unsafe integer cast: compile under integer context
+            v = int(e.as_rational())
+            return fpc.Ctx({ 'precision': 'integer' }, fpc.Integer(v))
+        raise FPCoreCompileError('cannot compile unrounded constant', e.val)
 
     def _visit_hexnum(self, e: Hexnum, ctx: None):
-        return fpc.Hexnum(e.val)
+        if e.is_integer() and self.unsafe_int_cast:
+            # unsafe integer cast: compile under integer context
+            v = int(e.as_rational())
+            return fpc.Ctx({ 'precision': 'integer' }, fpc.Integer(v))
+        raise FPCoreCompileError('cannot compile unrounded constant', e.val)
 
     def _visit_integer(self, e: Integer, ctx: None) -> fpc.Expr:
-        return fpc.Integer(e.val)
+        if self.unsafe_int_cast:
+            # unsafe integer cast: compile under integer context
+            return fpc.Ctx({ 'precision': 'integer' }, fpc.Integer(e.val))
+        else:
+            raise FPCoreCompileError('cannot compile unrounded constant', e.val)
 
     def _visit_rational(self, e: Rational, ctx: None):
-        return fpc.Rational(e.p, e.q)
-
-    def _visit_constant(self, e: Constant, ctx: None):
-        return fpc.Constant(e.val)
+        if self.unsafe_int_cast and e.is_integer():
+            # unsafe integer cast: compile under integer context
+            v = int(e.as_rational())
+            return fpc.Ctx({ 'precision': 'integer' }, fpc.Integer(v))
+        raise FPCoreCompileError('cannot compile unrounded constant', f'{e.p}/{e.q}')
 
     def _visit_digits(self, e: Digits, ctx: None) -> fpc.Expr:
-        return fpc.Digits(e.m, e.e, e.b)
+        if self.unsafe_int_cast and e.e == 0 and e.b == 2:
+            # unsafe integer cast: compile under integer context
+            v = int(e.as_rational())
+            return fpc.Ctx({ 'precision': 'integer' }, fpc.Integer(v))
+        raise FPCoreCompileError('cannot compile unrounded constant', f'digits({e.m}, {e.e}, {e.b})')
 
     def _visit_call(self, e: Call, ctx: None) -> fpc.Expr:
+        if e.kwargs:
+            raise FPCoreCompileError('cannot compile keyword arguments to FPCore', e)
+        match e.func:
+            case Var():
+                name = str(e.func.name)
+            case Attribute():
+                raise FPCoreCompileError('cannot compile method call to FPCore', e.func)
+            case _:
+                raise RuntimeError('unreachable', e.func)
         args = [self._visit_expr(c, ctx) for c in e.args]
-        return fpc.UnknownOperator(*args, name=e.name)
+        return fpc.UnknownOperator(*args, name=name)
+
+    def _visit_round(self, e: Round | RoundExact, ctx: None) -> fpc.Expr:
+        # round expression
+        match e.arg:
+            case Decnum():
+                # round(n) => n
+                return fpc.Decnum(e.arg.val)
+            case Hexnum():
+                # round(n) => n
+                return fpc.Hexnum(e.arg.val)
+            case Integer():
+                # round(n) => n
+                return fpc.Integer(e.arg.val)
+            case Rational():
+                # round(p/q) => p/q
+                return fpc.Rational(e.arg.p, e.arg.q)
+            case Digits():
+                # round(digits(m, e, b)) => digits(m, e, b)
+                return fpc.Digits(e.arg.m, e.arg.e, e.arg.b)
+            case _:
+                # round(e) => cast(e)
+                arg = self._visit_expr(e.arg, ctx)
+                return fpc.Cast(arg)
+
+    def _visit_round_at(self, e: RoundAt, ctx: None) -> fpc.Expr:
+        raise FPCoreCompileError('cannot compile `round_at` expression to FPCore', e)
+
+    def _visit_len(self, arg: Expr, ctx: None) -> fpc.Expr:
+        # length expression
+        arr = self._visit_expr(arg, ctx)
+        return fpc.Size(arr, fpc.Integer(0))
 
     def _visit_range(self, arg: Expr, ctx: None) -> fpc.Expr:
         # expand range expression
@@ -204,6 +308,12 @@ class FPCoreCompileInstance(Visitor):
         size = self._visit_expr(arg, ctx)
         return fpc.Tensor([(tuple_id, size)], fpc.Var(tuple_id))
 
+    def _visit_empty(self, arg: Expr, ctx: None) -> fpc.Expr:
+        # tensor with uninitialized values
+        tuple_id = str(self.gensym.fresh('i'))
+        size = self._visit_expr(arg, ctx)
+        return fpc.Tensor([(tuple_id, size)], fpc.Integer(0))
+
     def _visit_size(self, arr: Expr, dim: Expr, ctx) -> fpc.Expr:
         tup = self._visit_expr(arr, ctx)
         idx = self._visit_expr(dim, ctx)
@@ -213,28 +323,26 @@ class FPCoreCompileInstance(Visitor):
         tup = self._visit_expr(arr, ctx)
         return fpc.Dim(tup)
 
-    def _visit_shape(self, arr: Expr, ctx) -> fpc.Expr:
-        # expand into a for loop
-        #  (let ([t <tuple>])
-        #    (tensor ([i (dim t)])
-        #      (size t i)])))
-        tuple_id = str(self.gensym.fresh('t'))
+    def _visit_enumerate(self, arr: Expr, ctx: None) -> fpc.Expr:
+        # (let ([t <tuple>])
+        #  (tensor ([i (size t 0)])
+        #    (array i (ref t i))))
+        tuple_id = str(self.gensym.fresh('i'))
         iter_id = str(self.gensym.fresh('i'))
         tup = self._visit_expr(arr, ctx)
         return fpc.Let(
             [(tuple_id, tup)],
             fpc.Tensor(
-                [(iter_id, fpc.Dim(fpc.Var(tuple_id)))],
-                fpc.Size(fpc.Var(tuple_id), fpc.Var(iter_id))
+                [(iter_id, _size0_expr(tuple_id))],
+                fpc.Array(fpc.Var(iter_id), fpc.Ref(fpc.Var(tuple_id), fpc.Var(iter_id)))
             )
         )
 
-    def _visit_zip(self, args: list[Expr], ctx: None) -> fpc.Expr:
+    def _visit_zip(self, args: tuple[Expr, ...], ctx: None) -> fpc.Expr:
         # expand zip expression (for N=2)
         #  (let ([t0 <tuple0>] [t1 <tuple1>])
         #    (tensor ([i (size t0 0)])
         #      (array (ref t0 i) (ref t1 i)))))
-
         if len(args) == 0:
             # no children => empty zip
             return fpc.Array()
@@ -249,28 +357,93 @@ class FPCoreCompileInstance(Visitor):
                 )
             )
 
+    def _visit_min(self, args: tuple[Expr, ...], ctx: None) -> fpc.Expr:
+        # expand min expression by left associativity
+        # at least two arguments
+        # (fmin (fmin (fmin t1 t2) t3) ... tn)
+        vals = [self._visit_expr(arg, ctx) for arg in args]
+        min_expr = vals[0]
+        for val in vals[1:]:
+            min_expr = fpc.Fmin(min_expr, val)
+        return min_expr
+
+    def _visit_max(self, args: tuple[Expr, ...], ctx: None) -> fpc.Expr:
+        # expand max expression by left associativity
+        # at least two arguments
+        # (fmax (fmax (fmax t1 t2) t3) ... tn)
+        vals = [self._visit_expr(arg, ctx) for arg in args]
+        min_expr = vals[0]
+        for val in vals[1:]:
+            min_expr = fpc.Fmax(min_expr, val)
+        return min_expr
+
+    def _visit_sum(self, arg: Expr, ctx: None) -> fpc.Expr:
+        # expand sum expression by left associativity
+        # the sum expression has at most one argument
+        # (let ([t <tuple>])
+        #   (for ([i (! :precision integer (- (size t 0) 1))]
+        #         [accum (ref t i) (+ accum (ref t (! :precision integer (+ i 1))))])
+        #     accum))
+        tuple_id = str(self.gensym.fresh('t'))
+        iter_id = str(self.gensym.fresh('i'))
+        accum_id = str(self.gensym.fresh('accum'))
+        idx_ctx = { 'precision': 'integer' }
+
+        tup = self._visit_expr(arg, ctx)
+        return fpc.Let(
+            [(tuple_id, tup)],
+            fpc.For(
+                [(iter_id, fpc.Ctx(idx_ctx, fpc.Sub(fpc.Size(fpc.Var(tuple_id), fpc.Integer(0)), fpc.Integer(1))))],
+                [(
+                    accum_id,
+                    fpc.Ref(fpc.Var(tuple_id), fpc.Var(iter_id)),
+                    fpc.Ctx(idx_ctx, fpc.Add(accum_id, fpc.Ref(fpc.Var(tuple_id), fpc.Add(fpc.Var(iter_id), fpc.Integer(1)))))
+                )],
+                fpc.Var(accum_id)
+            )
+        )
+
+
+    def _visit_nullaryop(self, e: NullaryOp, ctx: None) -> fpc.Expr:
+        nullary_table = _get_nullary_table()
+        if type(e) not in nullary_table:
+            # unknown operator
+            raise NotImplementedError('no FPCore operator for', e)
+        return nullary_table[type(e)]
+
     def _visit_unaryop(self, e: UnaryOp, ctx: None) -> fpc.Expr:
-        cls = _unary_table.get(type(e))
+        unary_table = _get_unary_table()
+        cls = unary_table.get(type(e))
         if cls is not None:
             # known unary operator
             arg = self._visit_expr(e.arg, ctx)
             return cls(arg)
         else:
             match e:
+                case Len():
+                    # len expression
+                    return self._visit_len(e.arg, ctx)
                 case Range():
                     # range expression
                     return self._visit_range(e.arg, ctx)
+                case Empty():
+                    # empty expression
+                    return self._visit_empty(e.arg, ctx)
                 case Dim():
                     # dim expression
                     return self._visit_dim(e.arg, ctx)
-                case Shape():
-                    # shape expression
-                    return self._visit_shape(e.arg, ctx)
+                case Enumerate():
+                    # enumerate expression
+                    return self._visit_enumerate(e.arg, ctx)
+                case Sum():
+                    # sum expression
+                    return self._visit_sum(e.arg, ctx)
                 case _:
                     raise NotImplementedError('no FPCore operator for', e)
 
     def _visit_binaryop(self, e: BinaryOp, ctx: None) -> fpc.Expr:
-        cls = _binary_table.get(type(e))
+        binary_table = _get_binary_table()
+        cls = binary_table.get(type(e))
         if cls is not None:
             # known binary operator
             arg0 = self._visit_expr(e.first, ctx)
@@ -286,7 +459,8 @@ class FPCoreCompileInstance(Visitor):
                     raise NotImplementedError('no FPCore operator for', e)
 
     def _visit_ternaryop(self, e: TernaryOp, ctx: None) -> fpc.Expr:
-        cls = _ternary_table.get(type(e))
+        ternary_table = _get_ternary_table()
+        cls = ternary_table.get(type(e))
         if cls is not None:
             # known ternary operator
             arg0 = self._visit_expr(e.first, ctx)
@@ -298,7 +472,8 @@ class FPCoreCompileInstance(Visitor):
             raise NotImplementedError('no FPCore operator for', e)
 
     def _visit_naryop(self, e: NaryOp, ctx: None) -> fpc.Expr:
-        cls = _nary_table.get(type(e))
+        nary_table = _get_nary_table()
+        cls = nary_table.get(type(e))
         if cls is not None:
             # known n-ary operator
             return cls(*[self._visit_expr(c, ctx) for c in e.args])
@@ -307,9 +482,16 @@ class FPCoreCompileInstance(Visitor):
                 case Zip():
                     # zip expression
                     return self._visit_zip(e.args, ctx)
+                case Min():
+                    # min expression
+                    return self._visit_min(e.args, ctx)
+                case Max():
+                    # max expression
+                    return self._visit_max(e.args, ctx)
                 case _:
                     # unknown operator
                     raise NotImplementedError('no FPCore operator for', e)
+
     def _visit_compare(self, e: Compare, ctx: None) -> fpc.Expr:
         assert e.ops != [], 'should not be empty'
         match e.ops:
@@ -349,12 +531,85 @@ class FPCoreCompileInstance(Visitor):
                 raise NotImplementedError('unreachable', e.ops)
 
     def _visit_tuple_expr(self, e: TupleExpr, ctx: None) -> fpc.Expr:
-        return fpc.Array(*[self._visit_expr(c, ctx) for c in e.args])
+        return fpc.Array(*[self._visit_expr(c, ctx) for c in e.elts])
 
-    def _visit_tuple_ref(self, e: TupleRef, ctx: None) -> fpc.Expr:
+    def _visit_list_expr(self, e: ListExpr, ctx: None) -> fpc.Expr:
+        return fpc.Array(*[self._visit_expr(c, ctx) for c in e.elts])
+
+    def _visit_list_ref(self, e: ListRef, ctx: None) -> fpc.Expr:
+        t: Expr = e
+        indices: list[fpc.Expr] = []
+        while isinstance(t, ListRef):
+            index = self._visit_expr(t.index, ctx)
+            indices.append(index)
+            t = t.value
+
+        value = self._visit_expr(t, ctx)
+        return fpc.Ref(value, *reversed(indices))
+
+    def _visit_list_slice(self, e: ListSlice, ctx: None) -> fpc.Expr:
         value = self._visit_expr(e.value, ctx)
-        slices = [self._visit_expr(s, ctx) for s in e.slices]
-        return fpc.Ref(value, *slices)
+        match e.start, e.stop:
+            case None, None:
+                # produce a copy
+                # (let ([t <value>])
+                #   (tensor ([i (size t 0)]) (ref t i)))
+                tuple_id = str(self.gensym.fresh('t'))
+                iter_id = str(self.gensym.fresh('i'))
+                return fpc.Let(
+                    [(tuple_id, value)],
+                    fpc.Tensor([(iter_id, _size0_expr(tuple_id))],
+                        fpc.Ref(fpc.Var(tuple_id), fpc.Var(iter_id)))
+                )
+            case None, _:
+                # produce a truncated copy
+                # (let ([t <value>])
+                #   (tensor ([i <stop>]) (ref t i)))
+                assert isinstance(e.stop, Expr) # mypy doesn't like this match statement
+                tuple_id = str(self.gensym.fresh('t'))
+                iter_id = str(self.gensym.fresh('i'))
+                stop = self._visit_expr(e.stop, ctx)
+                return fpc.Let(
+                    [(tuple_id, value)],
+                    fpc.Tensor([(iter_id, stop)],
+                        fpc.Ref(fpc.Var(tuple_id), fpc.Var(iter_id)))
+                )
+            case _:
+                # default case
+                # (let ([t <value>]
+                #   (let ([len (size t 0)])
+                #     (let ([start (max 0 (min start len))]
+                #           [stop (max 0 (min stop len))])
+                #       (tensor ([i (fdim end start)]) (ref x (+ i start)))))
+                tuple_id = str(self.gensym.fresh('t'))
+                len_id = str(self.gensym.fresh('len'))
+                start_id = str(self.gensym.fresh('start'))
+                stop_id = str(self.gensym.fresh('stop'))
+                iter_id = str(self.gensym.fresh('i'))
+
+                assert e.start is not None
+                start = self._visit_expr(e.start, ctx)
+                if e.stop is None:
+                    stop = fpc.Size(fpc.Var(tuple_id), fpc.Integer(0))
+                else:
+                    stop = self._visit_expr(e.stop, ctx)
+
+                return fpc.Let(
+                    [(tuple_id, value)],
+                    fpc.Let(
+                        [(len_id, _size0_expr(tuple_id))],
+                        fpc.Let(
+                            [
+                                (start_id, fpc.Fmax(fpc.Integer(0), fpc.Fmin(start, fpc.Var(len_id)))),
+                                (stop_id, fpc.Fmax(fpc.Integer(0), fpc.Fmin(stop, fpc.Var(len_id))))
+                            ],
+                            fpc.Tensor(
+                                [(iter_id, fpc.Fdim(fpc.Var(stop_id), fpc.Var(start_id)))],
+                                fpc.Ref(fpc.Var(tuple_id), fpc.Add(fpc.Var(iter_id), fpc.Var(start_id)))
+                            )
+                        )
+                    )
+                )
 
     def _generate_tuple_set(self, tuple_id: str, iter_id: str, idx_ids: list[str], val_id: str):
         # dimension bindings
@@ -372,7 +627,7 @@ class FPCoreCompileInstance(Visitor):
         if_expr = fpc.If(cond_expr, ift_expr, iff_expr)
         return fpc.Tensor(tensor_dims, if_expr)
 
-    def _visit_tuple_set(self, e: TupleSet, ctx: None) -> fpc.Expr:
+    def _visit_list_set(self, e: ListSet, ctx: None) -> fpc.Expr:
         # general case:
         # 
         #   (let ([t <tuple>] [i0 <index>] ... [v <value>]))
@@ -398,14 +653,14 @@ class FPCoreCompileInstance(Visitor):
 
         # generate temporary variables
         tuple_id = str(self.gensym.fresh('t'))
-        idx_ids = [str(self.gensym.fresh('i')) for _ in e.slices]
+        idx_ids = [str(self.gensym.fresh('i')) for _ in e.indices]
         iter_id = str(self.gensym.fresh('k'))
         val_id = str(self.gensym.fresh('v'))
 
         # compile each component
-        tuple_expr = self._visit_expr(e.array, ctx)
-        idx_exprs = [self._visit_expr(idx, ctx) for idx in e.slices]
-        val_expr = self._visit_expr(e.value, ctx)
+        tuple_expr = self._visit_expr(e.value, ctx)
+        idx_exprs = [self._visit_expr(idx, ctx) for idx in e.indices]
+        val_expr = self._visit_expr(e.expr, ctx)
 
         # create initial let binding
         let_bindings = [(tuple_id, tuple_expr)]
@@ -418,7 +673,7 @@ class FPCoreCompileInstance(Visitor):
         return fpc.Let(let_bindings, tensor_expr)
 
 
-    def _visit_comp_expr(self, e: CompExpr, ctx: None) -> fpc.Expr:
+    def _visit_list_comp(self, e: ListComp, ctx: None) -> fpc.Expr:
         if len(e.targets) == 1:
             # simple case:
             # (let ([t <iterable>]) (tensor ([i (size t 0)]) (let ([<var> (ref t i)]) <elt>))
@@ -506,15 +761,18 @@ class FPCoreCompileInstance(Visitor):
         iff = self._visit_expr(e.iff, ctx)
         return fpc.If(cond, ift, iff)
 
+    def _visit_attribute(self, e: Attribute, ctx: None) -> fpc.Expr:
+        raise FPCoreCompileError(f'cannot compile to FPCore: {type(e).__name__}')
+
     def _visit_assign(self, stmt: Assign, ctx: fpc.Expr):
-        match stmt.binding:
+        match stmt.target:
             case Id():
-                bindings = [(str(stmt.binding), self._visit_expr(stmt.expr, None))]
+                bindings = [(str(stmt.target), self._visit_expr(stmt.expr, None))]
                 return fpc.Let(bindings, ctx)
             case TupleBinding():
                 tuple_id = str(self.gensym.fresh('t'))
                 tuple_bind = (tuple_id, self._visit_expr(stmt.expr, None))
-                destruct_bindings = self._compile_tuple_binding(tuple_id, stmt.binding, [])
+                destruct_bindings = self._compile_tuple_binding(tuple_id, stmt.target, [])
                 return fpc.LetStar([tuple_bind] + destruct_bindings, ctx)
             case _:
                 raise RuntimeError('unreachable', stmt.binding)
@@ -525,8 +783,7 @@ class FPCoreCompileInstance(Visitor):
     def _visit_if1(self, stmt: If1Stmt, ret: fpc.Expr):
         # check that only one variable is mutated in the loop
         # the `IfBundling` pass is required to ensure this
-        defs_in, defs_out = self.def_use.blocks[stmt.body]
-        mutated = defs_in.mutated_in(defs_out)
+        mutated = self.def_use.mutated_in(stmt.body)
         num_mutated = len(mutated)
 
         if num_mutated == 0:
@@ -554,16 +811,14 @@ class FPCoreCompileInstance(Visitor):
     def _visit_if(self, stmt: IfStmt, ret: fpc.Expr):
         # check that only one variable is mutated in the loop
         # the `IfBundling` pass is required to ensure this
-        defs_in_ift, defs_out_ift = self.def_use.blocks[stmt.ift]
-        defs_in_iff, defs_out_iff = self.def_use.blocks[stmt.iff]
-        mutated_ift = defs_in_ift.mutated_in(defs_out_ift)
-        mutated_iff = defs_in_iff.mutated_in(defs_out_iff)
-        mutated  = list(dict.fromkeys(mutated_ift + mutated_iff)) # union with ordering
+        mutated_ift = self.def_use.mutated_in(stmt.ift)
+        mutated_iff = self.def_use.mutated_in(stmt.iff)
+        mutated = sorted(mutated_ift | mutated_iff)
 
         # identify variables that were introduced in each body
-        intros_ift = defs_in_ift.fresh_in(defs_out_ift)
-        intros_iff = defs_in_iff.fresh_in(defs_out_iff)
-        intros = list(intros_ift & intros_iff) # intersection of fresh variables
+        intros_ift = self.def_use.introed_in(stmt.ift)
+        intros_iff = self.def_use.introed_in(stmt.iff)
+        intros = sorted(intros_ift & intros_iff) # intersection of fresh variables
 
         # mutated or introduced variables
         changed = mutated + intros
@@ -596,8 +851,7 @@ class FPCoreCompileInstance(Visitor):
     def _visit_while(self, stmt: WhileStmt, ret: fpc.Expr):
         # check that only one variable is mutated in the loop
         # the `WhileBundling` pass is required to ensure this
-        defs_in, defs_out = self.def_use.blocks[stmt.body]
-        mutated = defs_in.mutated_in(defs_out)
+        mutated = self.def_use.mutated_in(stmt.body)
         num_mutated = len(mutated)
 
         if num_mutated == 0:
@@ -622,8 +876,7 @@ class FPCoreCompileInstance(Visitor):
     def _visit_for(self, stmt: ForStmt, ret: fpc.Expr):
         # check that only one variable is mutated in the loop
         # the `ForBundling` pass is required to ensure this
-        defs_in, defs_out = self.def_use.blocks[stmt.body]
-        mutated = defs_in.mutated_in(defs_out)
+        mutated = self.def_use.mutated_in(stmt.body)
         num_mutated = len(mutated)
 
         if not isinstance(stmt.target, Id):
@@ -667,10 +920,6 @@ class FPCoreCompileInstance(Visitor):
                     ret
             ))
 
-
-    def _visit_context_expr(self, e: ContextExpr, ctx: None):
-        raise RuntimeError('do not call')
-
     def _visit_data(self, data):
         match data:
             case int():
@@ -685,15 +934,17 @@ class FPCoreCompileInstance(Visitor):
                 raise NotImplementedError(repr(data))
 
     def _visit_context(self, stmt: ContextStmt, ctx: None):
+        # check if the context is bound
+        if isinstance(stmt.target, NamedId):
+            raise FPCoreCompileError('Context statements cannot bind to a variable', stmt.target)
+
         body = self._visit_block(stmt.body, ctx)
         # extract a context value
         match stmt.ctx:
-            case ContextExpr() | Var():
-                raise FPCoreCompileError('Context expressions must be pre-computed', stmt.ctx)
             case ForeignVal():
                 val = stmt.ctx.val
             case _:
-                raise RuntimeError('unreachable', stmt.ctx)
+                raise FPCoreCompileError('Context expressions must be pre-computed', stmt.ctx)
 
         # convert to properties
         match val:
@@ -719,7 +970,10 @@ class FPCoreCompileInstance(Visitor):
     def _visit_return(self, stmt: ReturnStmt, ctx: None) -> fpc.Expr:
         return self._visit_expr(stmt.expr, ctx)
 
-    def _visit_block(self, block: StmtBlock, ctx: Optional[fpc.Expr]):
+    def _visit_pass(self, stmt: PassStmt, ctx: None) -> fpc.Expr:
+        return ctx
+
+    def _visit_block(self, block: StmtBlock, ctx: fpc.Expr | None):
         if ctx is None:
             e = self._visit_statement(block.stmts[-1], None)
             stmts = block.stmts[:-1]
@@ -734,12 +988,17 @@ class FPCoreCompileInstance(Visitor):
 
         return e
 
-    def _visit_function(self, func: FuncDef, ctx: Optional[fpc.Expr]):
+    def _visit_function(self, func: FuncDef, ctx: fpc.Expr | None):
         args = [self._compile_arg(arg) for arg in func.args]
         body = self._visit_block(func.body, ctx)
 
         # metadata
-        props = func.metadata.copy()
+        if func.meta is None:
+            props = {}
+        else:
+            props = func.meta.copy()
+
+        # context properties
         if func.ctx is not None:
             match func.ctx:
                 case Context():
@@ -780,11 +1039,19 @@ class FPCoreCompileInstance(Visitor):
         return super()._visit_statement(stmt, ctx)
 
 class FPCoreCompiler(Backend):
-    """Compiler from FPy IR to FPCore"""
+    """Compiler from FPy to FPCore"""
+
+    unsafe_int_cast: bool
+    """any unrounded integer is automatically compiled under an integer context"""
+
+    def __init__(self, *, unsafe_int_cast: bool = False):
+        self.unsafe_int_cast = unsafe_int_cast
+
+    def compile(self, func: Function, ctx: Context | None = None) -> fpc.FPCore:
+        # TODO: handle ctx
 
-    def compile(self, func: Function) -> fpc.FPCore:
         # normalization passes
-        ast = ContextInline.apply(func.ast, func.env)
+        ast = ConstFold.apply(func.ast, enable_op=False)
         ast = FuncUpdate.apply(ast)
         ast = ForUnpack.apply(ast)
         ast = ForBundling.apply(ast)
@@ -792,4 +1059,4 @@ class FPCoreCompiler(Backend):
         ast = IfBundling.apply(ast)
         # compile
         def_use = DefineUse.analyze(ast)
-        return FPCoreCompileInstance(ast, def_use).compile()
+        return _FPCoreCompileInstance(ast, def_use, self.unsafe_int_cast).compile()
