@@ -2,90 +2,136 @@
 FPy parsing from FPCore.
 """
 
-from typing import TypeAlias
+from typing import Any, TypeAlias
 
 import titanfp.fpbench.fpcast as fpc
 from titanfp.fpbench.fpcparser import data_as_expr
 
-from ..fpc_context import FPCoreContext, NoSuchContextError
 from ..ast.fpyast import *
-from ..analysis.syntax_check import SyntaxCheck
+from ..analysis import SyntaxCheck
+from ..env import ForeignEnv
+from ..fpc_context import FPCoreContext, NoSuchContextError
 from ..utils import Gensym, pythonize_id
 
 
 DataElt: TypeAlias = tuple['DataElt'] | fpc.ValueExpr
 
-_unary_table: dict[str, type[UnaryOp]] = {
-    'neg': Neg,
-    'not': Not,
-    'fabs': Fabs,
-    'sqrt': Sqrt,
-    'cbrt': Cbrt,
-    'ceil': Ceil,
-    'floor': Floor,
-    'nearbyint': NearbyInt,
-    'round': Round,
-    'trunc': Trunc,
-    'acos': Acos,
-    'asin': Asin,
-    'atan': Atan,
-    'cos': Cos,
-    'sin': Sin,
-    'tan': Tan,
-    'acosh': Acosh,
-    'asinh': Asinh,
-    'atanh': Atanh,
-    'cosh': Cosh,
-    'sinh': Sinh,
-    'tanh': Tanh,
-    'exp': Exp,
-    'exp2': Exp2,
-    'expm1': Expm1,
-    'log': Log,
-    'log10': Log10,
-    'log1p': Log1p,
-    'log2': Log2,
-    'erf': Erf,
-    'erfc': Erfc,
-    'lgamma': Lgamma,
-    'tgamma': Tgamma,
-    'isfinite': IsFinite,
-    'isinf': IsInf,
-    'isnan': IsNan,
-    'isnormal': IsNormal,
-    'signbit': Signbit,
-    'cast': Cast,
-    'range': Range,
-    'dim': Dim,
-}
+# Lazy lookup tables to avoid expensive initialization at import time
+_constants_cache = None
+_unary_table_cache = None
+_binary_table_cache = None
+_ternary_table_cache = None
 
-_binary_table: dict[str, type[BinaryOp]] = {
-    '+': Add,
-    '-': Sub,
-    '*': Mul,
-    '/': Div,
-    'copysign': Copysign,
-    'fdim': Fdim,
-    'fmax': Fmax,
-    'fmin': Fmin,
-    'fmod': Fmod,
-    'remainder': Remainder,
-    'hypot': Hypot,
-    'atan2': Atan2,
-    'pow': Pow,
-}
+def _get_constants():
+    global _constants_cache
+    if _constants_cache is None:
+        _constants_cache = {
+            'TRUE': BoolVal(True, None),
+            'FALSE': BoolVal(False, None),
+            'NAN': ConstNan(_func_symbol('nan'), None),
+            'INFINITY': ConstInf(_func_symbol('inf'), None),
+            'PI': ConstPi(_func_symbol('const_pi'), None),
+            'E': ConstE(_func_symbol('const_e'), None),
+            'LOG2E': ConstLog2E(_func_symbol('const_log2e'), None),
+            'LOG10E': ConstLog10E(_func_symbol('const_log10e'), None),
+            'LN2': ConstLn2(_func_symbol('const_ln2'), None),
+            'PI_2': ConstPi_2(_func_symbol('const_pi_2'), None),
+            'PI_4': ConstPi_4(_func_symbol('const_pi_4'), None),
+            'M_1_PI': Const1_Pi(_func_symbol('const_1_pi'), None),
+            'M_2_PI': Const2_Pi(_func_symbol('const_2_pi'), None),
+            'M_2_SQRTPI': Const2_SqrtPi(_func_symbol('const_2_sqrtpi'), None),
+            'SQRT2': ConstSqrt2(_func_symbol('const_sqrt2'), None),
+            'SQRT1_2': ConstSqrt1_2(_func_symbol('const_sqrt1_2'), None),
+        }
+    return _constants_cache
 
-_ternary_table: dict[str, type[TernaryOp]] = {
-    'fma': Fma
-}
+def _get_unary_table():
+    global _unary_table_cache
+    if _unary_table_cache is None:
+        _unary_table_cache = {
+            'neg': Neg,
+            'not': Not,
+            'fabs': Fabs,
+            'sqrt': Sqrt,
+            'cbrt': Cbrt,
+            'ceil': Ceil,
+            'floor': Floor,
+            'nearbyint': NearbyInt,
+            'round': RoundInt,
+            'trunc': Trunc,
+            'acos': Acos,
+            'asin': Asin,
+            'atan': Atan,
+            'cos': Cos,
+            'sin': Sin,
+            'tan': Tan,
+            'acosh': Acosh,
+            'asinh': Asinh,
+            'atanh': Atanh,
+            'cosh': Cosh,
+            'sinh': Sinh,
+            'tanh': Tanh,
+            'exp': Exp,
+            'exp2': Exp2,
+            'expm1': Expm1,
+            'log': Log,
+            'log10': Log10,
+            'log1p': Log1p,
+            'log2': Log2,
+            'erf': Erf,
+            'erfc': Erfc,
+            'lgamma': Lgamma,
+            'tgamma': Tgamma,
+            'isfinite': IsFinite,
+            'isinf': IsInf,
+            'isnan': IsNan,
+            'isnormal': IsNormal,
+            'signbit': Signbit,
+            'range': Range,
+            'dim': Dim,
+        }
+    return _unary_table_cache
 
-def _zeros(ns: list[Expr]) -> Expr:
-    assert len(ns) >= 1
-    v: Expr = Integer(0, None)
-    for n in reversed(ns):
-        v = CompExpr([UnderscoreId()], [Range(n, None)], v, None)
-    return v
+def _get_binary_table():
+    global _binary_table_cache
+    if _binary_table_cache is None:
+        _binary_table_cache = {
+            '+': Add,
+            '-': Sub,
+            '*': Mul,
+            '/': Div,
+            'copysign': Copysign,
+            'fdim': Fdim,
+            'fmod': Fmod,
+            'remainder': Remainder,
+            'hypot': Hypot,
+            'atan2': Atan2,
+            'pow': Pow,
+        }
+    return _binary_table_cache
 
+def _get_ternary_table():
+    global _ternary_table_cache
+    if _ternary_table_cache is None:
+        _ternary_table_cache = {
+            'fma': Fma
+        }
+    return _ternary_table_cache
+
+def _func_symbol(name: str):
+    return Var(NamedId(name), None)
+
+def _empty(ns: list[Expr]) -> Expr:
+    if len(ns) == 1:
+        return Empty(_func_symbol('empty'), ns[0], None)
+    elif len(ns) > 1:
+        elt = _empty(ns[1:])
+        return ListComp([UnderscoreId()], [Range(_func_symbol('range'), ns[0], None)], elt, None)
+    else:
+        raise ValueError(f'ns={ns} cannot be empty')
+
+def _round(x: Expr) -> Expr:
+    return Round(_func_symbol('round'), x, None)
 
 # TODO: clean this up
 class _Ctx:
@@ -95,9 +141,9 @@ class _Ctx:
 
     def __init__(
         self,
-        env: Optional[dict[str, NamedId]] = None,
-        props: Optional[dict[str, Any]] = None,
-        stmts: Optional[list[Stmt]] = None
+        env: dict[str, NamedId] | None = None,
+        props: dict[str, Any] | None = None,
+        stmts: list[Stmt] | None = None
     ):
         if env is None:
             self.env = {}
@@ -126,69 +172,92 @@ class _FPCore2FPy:
     core: fpc.FPCore
     gensym: Gensym
     default_name: str
+    env: ForeignEnv | None
 
-    def __init__(self, core: fpc.FPCore, default_name: str):
+    def __init__(self, core: fpc.FPCore, default_name: str, env: ForeignEnv | None = None):
         self.core = core
         self.gensym = Gensym(rename_hook=pythonize_id)
         self.default_name = default_name
+        self.env = env
 
     def _visit_var(self, e: fpc.Var, ctx: _Ctx) -> Expr:
         if e.value not in ctx.env:
             raise ValueError(f'variable {e.value} not in scope')
         return Var(ctx.env[e.value], None)
 
+    def _visit_constant(self, e: fpc.Constant, ctx: _Ctx) -> Expr:
+        constants = _get_constants()
+        if e.value not in constants:
+            raise ValueError(f'unknown constant {e.name}')
+        return constants[e.value]
+
     def _visit_decnum(self, e: fpc.Decnum, ctx: _Ctx) -> Expr:
-        return Decnum(str(e.value), None)
+        return _round(Decnum(str(e.value), None))
 
     def _visit_hexnum(self, e: fpc.Hexnum, ctx: _Ctx) -> Expr:
-        return Hexnum(str(e.value), None)
+        return _round(Hexnum(_func_symbol('hexnum'), str(e.value), None))
 
     def _visit_integer(self, e: fpc.Integer, ctx: _Ctx) -> Expr:
-        return Integer(int(e.value), None)
+        return _round(Integer(int(e.value), None))
 
     def _visit_rational(self, e: fpc.Rational, ctx: _Ctx) -> Expr:
-        return Rational(e.p, e.q, None)
+        return _round(Rational(_func_symbol('rational'), e.p, e.q, None))
 
     def _visit_digits(self, e: fpc.Digits, ctx: _Ctx) -> Expr:
-        return Digits(e.m, e.e, e.b, None)
-
-    def _visit_constant(self, e: fpc.Constant, ctx: _Ctx) -> Expr:
-        match e.value:
-            case 'TRUE':
-                return BoolVal(True, None)
-            case 'FALSE':
-                return BoolVal(False, None)
-            case _:
-                return Constant(str(e.value), None)
+        return _round(Digits(_func_symbol('digits'), e.m, e.e, e.b, None))
 
     def _visit_unary(self, e: fpc.UnaryExpr, ctx: _Ctx) -> Expr:
+        unary_table = _get_unary_table()
         if e.name == '-':
-            cls = _unary_table['neg']
             arg = self._visit(e.children[0], ctx)
-            return cls(arg, None)
-        elif e.name in _unary_table:
-            cls = _unary_table[e.name]
+            return Neg(arg, None)
+        elif e.name == 'cast':
             arg = self._visit(e.children[0], ctx)
-            return cls(arg, None)
+            return _round(arg)
+        elif e.name in unary_table:
+            cls = unary_table[e.name]
+            arg = self._visit(e.children[0], ctx)
+            if issubclass(cls, NamedUnaryOp):
+                return cls(_func_symbol(e.name), arg, None)
+            else:
+                return cls(arg, None)
         else:
             raise NotImplementedError(f'unsupported unary operation {e.name}')
 
     def _visit_binary(self, e: fpc.BinaryExpr, ctx: _Ctx) -> Expr:
-        if e.name in _binary_table:
-            cls = _binary_table[e.name]
+        binary_table = _get_binary_table()
+        if e.name in binary_table:
+            cls = binary_table[e.name]
             left = self._visit(e.children[0], ctx)
             right = self._visit(e.children[1], ctx)
-            return cls(left, right, None)
+            if issubclass(cls, NamedBinaryOp):
+                return cls(_func_symbol(e.name), left, right, None)
+            else:
+                return cls(left, right, None)
         else:
-            raise NotImplementedError(f'unsupported binary operation {e.name}')
+            match e.name:
+                case 'fmin':
+                    left = self._visit(e.children[0], ctx)
+                    right = self._visit(e.children[1], ctx)
+                    return Min(_func_symbol('min'), (left, right), None)
+                case 'fmax':
+                    left = self._visit(e.children[0], ctx)
+                    right = self._visit(e.children[1], ctx)
+                    return Max(_func_symbol('max'), (left, right), None)
+                case _:
+                    raise NotImplementedError(f'unsupported binary operation {e.name}')
 
     def _visit_ternary(self, e: fpc.TernaryExpr, ctx: _Ctx) -> Expr:
-        if e.name in _ternary_table:
-            cls = _ternary_table[e.name]
+        ternary_table = _get_ternary_table()
+        if e.name in ternary_table:
+            cls = ternary_table[e.name]
             arg0 = self._visit(e.children[0], ctx)
             arg1 = self._visit(e.children[1], ctx)
             arg2 = self._visit(e.children[2], ctx)
-            return cls(arg0, arg1, arg2, None)
+            if issubclass(cls, NamedTernaryOp):
+                return cls(_func_symbol(e.name), arg0, arg1, arg2, None)
+            else:
+                return cls(arg0, arg1, arg2, None)
         else:
             raise NotImplementedError(f'unsupported ternary operation {e.name}')
 
@@ -237,22 +306,28 @@ class _FPCore2FPy:
                     raise ValueError('size operator expects 2 arguments')
                 arg0 = self._visit(e.children[0], ctx)
                 arg1 = self._visit(e.children[1], ctx)
-                return Size(arg0, arg1, None)
+                return Size(_func_symbol('size'), arg0, arg1, None)
             case fpc.UnknownOperator():
-                name = pythonize_id(e.name)
+                ident = pythonize_id(e.name)
                 exprs = [self._visit(e, ctx) for e in e.children]
-                return Call(name, exprs, None)
+                return Call(_func_symbol(ident), None, exprs, {}, None)
             case _:
                 raise NotImplementedError('unexpected FPCore expression', e)
 
     def _visit_array(self, e: fpc.Array, ctx: _Ctx) -> Expr:
         exprs = [self._visit(e, ctx) for e in e.children]
-        return TupleExpr(exprs, None)
+        return ListExpr(exprs, None)
 
     def _visit_ref(self, e: fpc.Ref, ctx: _Ctx) -> Expr:
-        value = self._visit(e.children[0], ctx)
-        slices = [self._visit(e, ctx) for e in e.children[1:]]
-        return TupleRef(value, slices, None)
+        # (ref <array> <index> ...)
+        # => <array>[<index>][...]
+        arr = self._visit(e.children[0], ctx)
+        indices = [self._visit(e, ctx) for e in e.children[1:]]
+
+        val: fpc.Expr = arr
+        for index in indices:
+            val = ListRef(val, index, None)
+        return val
 
     def _visit_if(self, e: fpc.If, ctx: _Ctx) -> Expr:
         # create new blocks
@@ -380,7 +455,7 @@ class _FPCore2FPy:
             t = iter_vars[0]
             n = range_vars[0]
             inner_stmts: list[Stmt] = []
-            e = Range(Var(n, None), None)
+            e = Range(_func_symbol('range'), Var(n, None), None)
             stmt = ForStmt(t, e, StmtBlock(inner_stmts), None)
             stmts.append(stmt)
             return self._make_tensor_body(iter_vars[1:], range_vars[1:], inner_stmts)
@@ -422,7 +497,7 @@ class _FPCore2FPy:
 
         # initialize tensor
         tuple_id = self.gensym.fresh('t')
-        zeroed = _zeros([Var(var, None) for var in bound_vars])
+        zeroed = _empty([Var(var, None) for var in bound_vars])
         stmt = Assign(tuple_id, None, zeroed, None)
         ctx.stmts.append(stmt)
 
@@ -475,7 +550,7 @@ class _FPCore2FPy:
 
         # initialize tensor
         tuple_id = self.gensym.fresh('t')
-        zeroed = _zeros([Var(var, None) for var in bound_vars])
+        zeroed = _empty([Var(var, None) for var in bound_vars])
         stmt = Assign(tuple_id, None, zeroed, None)
         ctx.stmts.append(stmt)
 
@@ -601,6 +676,8 @@ class _FPCore2FPy:
         match e:
             case fpc.Var():
                 return self._visit_var(e, ctx)
+            case fpc.Constant():
+                return self._visit_constant(e, ctx)
             case fpc.Decnum():
                 return self._visit_decnum(e, ctx)
             case fpc.Hexnum():
@@ -611,8 +688,6 @@ class _FPCore2FPy:
                 return self._visit_rational(e, ctx)
             case fpc.Digits():
                 return self._visit_digits(e, ctx)
-            case fpc.Constant():
-                return self._visit_constant(e, ctx)
             case fpc.UnaryExpr():
                 return self._visit_unary(e, ctx)
             case fpc.BinaryExpr():
@@ -673,33 +748,34 @@ class _FPCore2FPy:
         # compile arguments
         args: list[Argument] = []
         for name, _, shape in f.inputs:
+            # argument id
             t = self.gensym.fresh(name)
             match shape:
                 case tuple() | list():
                     # tensor argument
+                    # bind named tensor dimensions
                     dims: list[int | NamedId] = []
-                    dim_ids: list[Id] = []
-                    for dim in shape:
-                        if isinstance(dim, str):
-                            d = self.gensym.fresh(dim)
-                            dims.append(d)
-                            dim_ids.append(d)
-                        else:
-                            dims.append(dim)
-                            dim_ids.append(UnderscoreId())
+                    for i, dim in enumerate(shape):
+                        match dim:
+                            case int():
+                                # definite size dimension
+                                dims.append(dim)
+                            case str():
+                                # named dimension
+                                dim_id = self.gensym.fresh(dim)
+                                size_e = Size(_func_symbol('size'), Var(t, None), Integer(i, None), None)
+                                stmt = Assign(dim_id, None, size_e, None)
+                                ctx.stmts.append(stmt)
+                                # TODO: duplicate dimension names means a runtime check
+                                # How should this be expressed in FPy?
+                                ctx.env[dim] = dim_id
+                                dims.append(dim_id)
+                            case _:
+                                raise RuntimeError(f'unsupported shape {shape} for argument {name}')
 
                     arg = Argument(t, SizedTensorTypeAnn(dims, AnyTypeAnn(None), None), None)
                     args.append(arg)
                     ctx.env[name] = t
-
-                    # bind tensor dimensions
-                    shape_e = Shape(Var(t, None), None)
-                    stmt = Assign(TupleBinding(dim_ids, None), None, shape_e, None)
-                    ctx.stmts.append(stmt)
-                    for dim, dim_id in zip(shape, dim_ids):
-                        if isinstance(dim, str):
-                            assert isinstance(dim_id, NamedId), "must be a NamedId"
-                            ctx.env[dim] = dim_id
                 case None:
                     # scalar argument
                     arg = Argument(t, None, None)
@@ -727,7 +803,8 @@ class _FPCore2FPy:
         block = StmtBlock(ctx.stmts + [ReturnStmt(e, None)])
 
         name = self.default_name if f.ident is None else pythonize_id(f.ident)
-        return FuncDef(name, args, block, metadata=props, ctx=ctx_val)
+        env = ForeignEnv.default() if self.env is None else self.env
+        return FuncDef(name, args, set(), ctx_val, block, None, props, env)
 
     def convert(self) -> FuncDef:
         return self._visit_function(self.core)
@@ -736,9 +813,12 @@ class _FPCore2FPy:
 def fpcore_to_fpy(
     core: fpc.FPCore,
     *,
+    env: ForeignEnv | None = None,
     default_name: str = 'f',
     ignore_unknown: bool = False
 ):
-    ast = _FPCore2FPy(core, default_name).convert()
+    # TODO: support `prefix` argument to list how
+    # FPy builtins are printed
+    ast = _FPCore2FPy(core, default_name, env).convert()
     SyntaxCheck.check(ast, ignore_unknown=ignore_unknown)
     return ast
