@@ -4,134 +4,122 @@ This module contains the parser for the FPy language.
 
 import ast
 
-from typing import Mapping
+from typing import Any, Callable, Mapping
 from types import FunctionType
 
 from ..ast.fpyast import *
+from ..env import ForeignEnv
+from ..number import Float, Real
 from ..utils import NamedId, UnderscoreId, SourceId
+from ..ops import *
 
-def _ipow(expr: Expr, n: int, loc: Location):
-    assert n >= 0, "must be a non-negative integer"
-    if n == 0:
-        return Integer(1, loc)
-    elif n == 1:
-        return expr
-    else:
-        e = Mul(expr, expr, loc)
-        for _ in range(2, n):
-            e = Mul(e, expr, loc)
-        return e
-
-_constants: set[str] = {
-    'PI',
-    'E',
-    'NAN',
-    'INFINITY',
-    'LOG2E',
-    'LOG10E',
-    'LN2',
-    'PI_2',
-    'PI_4',
-    'M_1_PI',
-    'M_2_PI',
-    'M_2_SQRTPI',
-    'SQRT2',
-    'SQRT1_2'
+_nullary_table: dict[Callable, type[NullaryOp]] = {
+    nan: ConstNan,
+    inf: ConstInf,
+    const_pi: ConstPi,
+    const_e: ConstE,
+    const_log2e: ConstLog2E,
+    const_log10e: ConstLog10E,
+    const_ln2: ConstLn2,
+    const_pi_2: ConstPi_2,
+    const_pi_4: ConstPi_4,
+    const_1_pi: Const1_Pi,
+    const_2_pi: Const2_Pi,
+    const_2_sqrt_pi: Const2_SqrtPi,
+    const_sqrt2: ConstSqrt2,
+    const_sqrt1_2: ConstSqrt1_2,
 }
 
-_unary_table: dict[str, type[UnaryOp]] = {
-    'abs': Fabs,
-    'fabs': Fabs,
-    'sqrt': Sqrt,
-    'cbrt': Cbrt,
-    'ceil': Ceil,
-    'floor': Floor,
-    'nearbyint': NearbyInt,
-    'round': Round,
-    'trunc': Trunc,
-    'acos': Acos,
-    'asin': Asin,
-    'atan': Atan,
-    'cos': Cos,
-    'sin': Sin,
-    'tan': Tan,
-    'acosh': Acosh,
-    'asinh': Asinh,
-    'atanh': Atanh,
-    'cosh': Cosh,
-    'sinh': Sinh,
-    'tanh': Tanh,
-    'exp': Exp,
-    'exp2': Exp2,
-    'expm1': Expm1,
-    'log': Log,
-    'log10': Log10,
-    'log1p': Log1p,
-    'log2': Log2,
-    'erf': Erf,
-    'erfc': Erfc,
-    'lgamma': Lgamma,
-    'tgamma': Tgamma,
-    'isfinite': IsFinite,
-    'isinf': IsInf,
-    'isnan': IsNan,
-    'isnormal': IsNormal,
-    'signbit': Signbit,
-    'not': Not,
-    'cast': Cast,
-    'range': Range,
-    'shape': Shape,
-    'dim': Dim,
+_unary_table: dict[Callable, type[UnaryOp] | type[NamedUnaryOp]] = {
+    abs: Fabs,
+    fabs: Fabs,
+    sqrt: Sqrt,
+    cbrt: Cbrt,
+    ceil: Ceil,
+    floor: Floor,
+    nearbyint: NearbyInt,
+    roundint: RoundInt,
+    trunc: Trunc,
+    acos: Acos,
+    asin: Asin,
+    atan: Atan,
+    cos: Cos,
+    sin: Sin,
+    tan: Tan,
+    acosh: Acosh,
+    asinh: Asinh,
+    atanh: Atanh,
+    cosh: Cosh,
+    sinh: Sinh,
+    tanh: Tanh,
+    exp: Exp,
+    exp2: Exp2,
+    expm1: Expm1,
+    log: Log,
+    log10: Log10,
+    log1p: Log1p,
+    log2: Log2,
+    erf: Erf,
+    erfc: Erfc,
+    lgamma: Lgamma,
+    tgamma: Tgamma,
+    isfinite: IsFinite,
+    isinf: IsInf,
+    isnan: IsNan,
+    isnormal: IsNormal,
+    signbit: Signbit,
+    round: Round,
+    round_exact: RoundExact,
+    len: Len,
+    range: Range,
+    empty: Empty,
+    dim: Dim,
+    enumerate: Enumerate,
+    sum: Sum
 }
 
-_binary_table: dict[str, type[BinaryOp]] = {
-    'add': Add,
-    'sub': Sub,
-    'mul': Mul,
-    'div': Div,
-    'copysign': Copysign,
-    'fdim': Fdim,
-    'fmax': Fmax,
-    'min': Fmin,
-    'max': Fmax,
-    'fmin': Fmin,
-    'fmod': Fmod,
-    'remainder': Remainder,
-    'hypot': Hypot,
-    'atan2': Atan2,
-    'pow': Pow,
-    'size': Size
+_binary_table: dict[Callable, type[BinaryOp] | type[NamedBinaryOp]] = {
+    add: Add,
+    sub: Sub,
+    mul: Mul,
+    div: Div,
+    copysign: Copysign,
+    fdim: Fdim,
+    fmod: Fmod,
+    remainder: Remainder,
+    hypot: Hypot,
+    atan2: Atan2,
+    pow: Pow,
+    size: Size,
+    round_at: RoundAt,
 }
 
-_ternary_table: dict[str, type[TernaryOp]] = {
-    'fma': Fma
+_ternary_table: dict[Callable, type[TernaryOp] | type[NamedTernaryOp]] = {
+    fma: Fma
 }
 
-_nary_table: dict[str, type[NaryOp]] = {
-    'and': And,
-    'or': Or,
-    'zip': Zip,
+_nary_table: dict[Callable, type[NaryOp] | type[NamedNaryOp]] = {
+    zip: Zip,
+    max: Max,
+    min: Min,
+    fmin: Min,
+    fmax: Max,
 }
 
-_special_functions = {
-    'digits',
-    'hexfloat',
-    'rational'
-}
 
 class FPyParserError(Exception):
     """Parser error for FPy"""
     loc: Location
     why: str
     where: ast.AST
-    ctx: Optional[ast.AST]
+    ctx: ast.AST | None
 
     def __init__(
         self,
         loc: Location,
         why: str,
         where: ast.AST,
-        ctx: Optional[ast.AST] = None
+        ctx: ast.AST | None = None
     ):
         msg_lines = [why]
         match where:
@@ -153,6 +141,18 @@ class FPyParserError(Exception):
         self.ctx = ctx
 
 
+def _ipow(expr: Expr, n: int, loc: Location):
+    assert n >= 0, "must be a non-negative integer"
+    if n == 0:
+        return Integer(1, loc)
+    elif n == 1:
+        return expr
+    else:
+        e = Mul(expr, expr, loc)
+        for _ in range(2, n):
+            e = Mul(e, expr, loc)
+        return e
+
 class Parser:
     """
     FPy parser.
@@ -162,6 +162,7 @@ class Parser:
 
     name: str
     source: str
+    env: ForeignEnv
     lines: list[str]
     start_line: int
 
@@ -169,27 +170,15 @@ class Parser:
         self,
         name: str, 
         source: str,
+        env: ForeignEnv,
         start_line: int = 1
     ):
         self.name = name
         self.source = source
+        self.env = env
         self.lines = source.splitlines()
         self.start_line = start_line
 
-    def parse_function(self):
-        """Parses `self.source` as an FPy `FunctionDef`."""
-        start_loc = Location(self.name, self.start_line, 0, self.start_line, 0)
-
-        mod = ast.parse(self.source, self.name)
-        if len(mod.body) > 1:
-            raise FPyParserError(start_loc, 'FPy only supports single function definitions', mod)
-
-        ptree = mod.body[0]
-        if not isinstance(ptree, ast.FunctionDef):
-            raise FPyParserError(start_loc, 'FPy only supports single function definitions', mod)
-
-        return self._parse_function(ptree)
-
     def _parse_location(self, e: ast.expr | ast.stmt | ast.arg) -> Location:
         """Extracts the parse location of a  Python ST node."""
         assert e.end_lineno is not None, "missing end line number"
@@ -202,22 +191,71 @@ class Parser:
             e.end_col_offset
         )
 
+    def _convert_type(self, ty, loc: Location):
+        if ty == Real:
+            return RealTypeAnn(loc)
+        elif isinstance(ty, type):
+            if issubclass(ty, bool):
+                return BoolTypeAnn(loc)
+            elif issubclass(ty, int) or issubclass(ty, float):
+                # TODO: more specific type
+                return RealTypeAnn(loc)
+            elif issubclass(ty, Float):
+                return RealTypeAnn(loc)
+            else:
+                # TODO: implement
+                return AnyTypeAnn(loc)
+        elif isinstance(ty, tuple):
+            elts = [self._convert_type(elt, loc) for elt in ty]
+            return TupleTypeAnn(elts, loc)
+        elif isinstance(ty, list):
+            elt = self._convert_type(ty[0], loc)
+            return ListTypeAnn(elt, loc)
+        else:
+            # TODO: implement
+            return AnyTypeAnn(loc)
+
+    def _eval_type_annotation(self, ann: ast.expr):
+        loc = self._parse_location(ann)
+        match ann:
+            case ast.Attribute():
+                attr = self._parse_attribute(ann)
+                return self._eval_attribute(attr, ann)
+            case ast.Name():
+                ident = self._parse_id(ann)
+                if isinstance(ident, UnderscoreId):
+                    raise FPyParserError(loc, 'FPy function call must begin with a named identifier', ann)
+                if ident.base not in self.env:
+                    raise FPyParserError(loc, f'name \'{ident.base}\' not defined:', ann)
+                return self.env[ident.base]
+            case ast.Subscript():
+                ctor = self._eval_type_annotation(ann.value)
+                if ctor is tuple:
+                    # tuple[t1, ...]
+                    arg = self._eval_type_annotation(ann.slice)
+                    match arg:
+                        case tuple():
+                            return arg
+                        case _:
+                            return (arg,)
+                elif ctor is list:
+                    # list[t]
+                    arg = self._eval_type_annotation(ann.slice)
+                    return [arg]
+                else:
+                    return None
+            case ast.Tuple():
+                return tuple(self._eval_type_annotation(elt) for elt in ann.elts)
+            case _:
+                # TODO: implement
+                return None
 
     def _parse_type_annotation(self, ann: ast.expr) -> TypeAnn:
         loc = self._parse_location(ann)
-        match ann:
-            case ast.Name('Real'):
-                return RealTypeAnn(loc)
-            case ast.Name('int'):
-                # TODO: more specific type
-                return RealTypeAnn(loc)
-            case ast.Name('float'):
-                return RealTypeAnn(loc)
-            case ast.Name('bool'):
-                return RealTypeAnn(loc)
-            case _:
-                # TODO: implement
-                return AnyTypeAnn(loc)
+        ty = self._eval_type_annotation(ann)
+        if ty is None:
+            return AnyTypeAnn(loc)
+        return self._convert_type(ty, loc)
 
     def _parse_id(self, e: ast.Name):
         if e.id == '_':
@@ -241,19 +279,21 @@ class Parser:
                     return Decnum(str(e.value), loc)
             case str():
                 return ForeignVal(e.value, loc)
+            case None:
+                return ForeignVal(e.value, loc)
             case _:
                 raise FPyParserError(loc, 'Unsupported constant', e)
 
-    def _parse_hexfloat(self, e: ast.Call):
+    def _parse_hexfloat(self, e: ast.Call, func: FuncSymbol):
         loc = self._parse_location(e)
         if len(e.args) != 1:
             raise FPyParserError(loc, 'FPy `hexfloat` expects one argument', e)
         arg = self._parse_expr(e.args[0])
         if not isinstance(arg, ForeignVal):
             raise FPyParserError(loc, 'FPy `hexfloat` expects a string', e)
-        return Hexnum(arg.val, loc)
+        return Hexnum(func, arg.val, loc)
 
-    def _parse_rational(self, e: ast.Call):
+    def _parse_rational(self, e: ast.Call, func: FuncSymbol):
         loc = self._parse_location(e)
         if len(e.args) != 2:
             raise FPyParserError(loc, 'FPy `rational` expects two arguments', e)
@@ -263,9 +303,9 @@ class Parser:
         q = self._parse_expr(e.args[1])
         if not isinstance(q, Integer):
             raise FPyParserError(loc, 'FPy `rational` expects an integer as second argument', e)
-        return Rational(p.val, q.val, loc)
+        return Rational(func, p.val, q.val, loc)
 
-    def _parse_digits(self, e: ast.Call):
+    def _parse_digits(self, e: ast.Call, func: FuncSymbol):
         loc = self._parse_location(e)
         if len(e.args) != 3:
             raise FPyParserError(loc, 'FPy `digits` expects three arguments', e)
@@ -278,7 +318,7 @@ class Parser:
         b_e = self._parse_expr(e.args[2])
         if not isinstance(b_e, Integer):
             raise FPyParserError(loc, 'FPy `digits` expects an integer as third argument', e)
-        return Digits(m_e.val, e_e.val, b_e.val, loc)
+        return Digits(func, m_e.val, e_e.val, b_e.val, loc)
 
     def _parse_boolop(self, e: ast.BoolOp):
         loc = self._parse_location(e)
@@ -361,31 +401,169 @@ class Parser:
         args = [self._parse_expr(e) for e in [e.left, *e.comparators]]
         return Compare(ops, args, loc)
 
-    def _parse_call(self, e: ast.Call) -> str:
-        """Parse a Python call expression."""
-        loc = self._parse_location(e)
-        if not isinstance(e.func, ast.Name):
-            raise FPyParserError(loc, 'Unsupported call expression', e)
-        return e.func.id
+    def _eval_var(self, v: Var, e: ast.expr):
+        if v.name.base not in self.env:
+            loc = self._parse_location(e)
+            raise FPyParserError(loc, f'name \'{v.name}\' not defined:', e)
+        return self.env[v.name.base]
 
-    def _parse_slice(self, slice: ast.expr, e: ast.expr):
-        match slice:
-            case ast.slice():
-                loc = self._parse_location(e)
-                raise  FPyParserError(loc, 'Slices unsupported', e, slice)
-            case ast.Tuple():
-                return [self._parse_expr(s) for s in slice.elts]
+    def _eval_attribute(self, a: Attribute, e: ast.expr):
+        match a.value:
+            case Var():
+                # evaluating `x.y`
+                base = self._eval_var(a.value, e)
+            case Attribute():
+                # evaluating `x.y.z` where `x.y` is `a`
+                base = self._eval_attribute(a.value, e)
+
+        # lookup the attribute
+        if not hasattr(base, a.attr):
+            loc = self._parse_location(e)
+            raise FPyParserError(loc, f'unknown attribute \'{a.attr}\' for {base}', e)
+
+        return getattr(base, a.attr)
+
+    def _parse_call(self, e: ast.Call):
+        """Parse a Python call function."""
+        # parse function expression
+        loc = self._parse_location(e.func)
+        match e.func:
+            case ast.Attribute():
+                func = self._parse_attribute(e.func)
+                fn = self._eval_attribute(func, e.func)
+            case ast.Name():
+                name = self._parse_id(e.func)
+                func = Var(name, None)
+                if isinstance(func, UnderscoreId):
+                    raise FPyParserError(loc, 'FPy function call must begin with a named identifier', e)
+                if name.base not in self.env:
+                    raise FPyParserError(loc, f'name \'{name.base}\' not defined:', e)
+                fn = self.env[name.base]
             case _:
-                return [self._parse_expr(slice)]
+                raise RuntimeError('unreachable')
+
+        # parse arguments
+        args = [self._parse_expr(arg) for arg in e.args]
+
+        # parse keyword arguments
+        kwargs: list[tuple[str, Expr]] = []
+        for kwarg in e.keywords:
+            if kwarg.arg is None:
+                raise FPyParserError(loc, 'FPy does not support **kwargs', e)
+            kwarg_val = self._parse_expr(kwarg.value)
+            kwargs.append((kwarg.arg, kwarg_val))
+
+        # lookup builtin symbols
+        if fn in _nullary_table:
+            cls0 = _nullary_table[fn]
+            if len(e.args) != 0:
+                raise FPyParserError(loc, f'FPy expects 0 arguments for `{fn}`, got {len(e.args)}', e)
+            if kwargs:
+                raise FPyParserError(loc, f'FPy does not support keyword arguments for `{fn}`', e)
+            return cls0(func, loc)
+        elif fn in _unary_table:
+            cls1 = _unary_table[fn]
+            if len(args) != 1:
+                raise FPyParserError(loc, f'FPy expects 1 argument for `{fn}`, got {len(e.args)}', e)
+            if kwargs:
+                raise FPyParserError(loc, f'FPy does not support keyword arguments for `{fn}`', e)
+            if issubclass(cls1, NamedUnaryOp):
+                return cls1(func, args[0], loc)
+            else:
+                return cls1(args[0], loc)
+        elif fn in _binary_table:
+            cls2 = _binary_table[fn]
+            if len(args) != 2:
+                raise FPyParserError(loc, f'FPy expects 2 arguments for `{fn}`, got {len(e.args)}', e)
+            if kwargs:
+                raise FPyParserError(loc, f'FPy does not support keyword arguments for `{fn}`', e)
+            if issubclass(cls2, NamedBinaryOp):
+                return cls2(func, args[0], args[1], loc)
+            else:
+                return cls2(args[0], args[1], loc)
+        elif fn in _ternary_table:
+            cls3 = _ternary_table[fn]
+            if len(args) != 3:
+                raise FPyParserError(loc, f'FPy expects 3 arguments for `{fn}`, got {len(e.args)}', e)
+            if kwargs:
+                raise FPyParserError(loc, f'FPy does not support keyword arguments for `{fn}`', e)
+            if issubclass(cls3, NamedTernaryOp):
+                return cls3(func, args[0], args[1], args[2], loc)
+            else:
+                return cls3(args[0], args[1], args[2], loc)
+        elif fn in _nary_table:
+            cls = _nary_table[fn]
+            if (cls is Min or cls is Max) and len(args) < 2:
+                raise FPyParserError(loc, f'FPy expects at least 2 arguments for `{fn}`', e)
+            if kwargs:
+                raise FPyParserError(loc, f'FPy does not support keyword arguments for `{fn}`', e)
+            if issubclass(cls, NamedNaryOp):
+                return cls(func, args, loc)
+            else:
+                return cls(args, loc)
+        elif fn == rational:
+            if kwargs:
+                raise FPyParserError(loc, f'FPy does not support keyword arguments for `{fn}`', e)
+            return self._parse_rational(e, func)
+        elif fn == hexfloat:
+            if kwargs:
+                raise FPyParserError(loc, f'FPy does not support keyword arguments for `{fn}`', e)
+            return self._parse_hexfloat(e, func)
+        elif fn == digits:
+            if kwargs:
+                raise FPyParserError(loc, f'FPy does not support keyword arguments for `{fn}`', e)
+            return self._parse_digits(e, func)
+        elif fn == len:
+            if len(args) != 1:
+                raise FPyParserError(loc, 'FPy expects 1 argument for `len`', e)
+            if kwargs:
+                raise FPyParserError(loc, 'FPy does not support keyword arguments for `len`', e)
+            return Size(func, args[0], Integer(0, None), loc)
+        else:
+            return Call(func, fn, args, kwargs, loc)
+
+    def _parse_slice(self, e: ast.Slice):
+        """Parse a Python slice expression."""
+        if e.lower is None:
+            lower = None
+        else:
+            lower = self._parse_expr(e.lower)
+
+        if e.upper is None:
+            upper = None
+        else:
+            upper = self._parse_expr(e.upper)
+
+        if e.step is not None:
+            loc = self._parse_location(e.step)
+            raise FPyParserError(loc, 'FPy does not support slice step', e.step)
+
+        return lower, upper
 
     def _parse_subscript(self, e: ast.Subscript):
+        """Parsing a subscript slice that is an expression"""
+        loc = self._parse_location(e)
         value = self._parse_expr(e.value)
-        slices = self._parse_slice(e.slice, e)
-        while isinstance(value, TupleRef):
-            v_value, v_slices = value.value, value.slices
-            value = v_value
-            slices = v_slices + slices
-        return (value, slices)
+        match e.slice:
+            case ast.Slice():
+                lower, upper = self._parse_slice(e.slice)
+                return ListSlice(value, lower, upper, loc)
+            case _:
+                slice =  self._parse_expr(e.slice)
+                return ListRef(value, slice, loc)
+
+    def _parse_subscript_target(self, e: ast.Subscript):
+        """Parsing a subscript slice that is the LHS of an assignment."""
+        t: ast.expr = e
+        slices: list[Expr] = []
+        while isinstance(t, ast.Subscript):
+            slices.append(self._parse_expr(t.slice))
+            t = t.value
+
+        target = self._parse_expr(t)
+        slices.reverse()
+
+        return target, slices
 
     def _is_foreign_val(self, e: ast.expr):
         match e:
@@ -396,34 +574,18 @@ class Parser:
             case _:
                 return False
 
-    def _parse_foreign_attribute(self, e: ast.expr):
+    def _parse_attribute(self, e: ast.Attribute):
         loc = self._parse_location(e)
-        match e:
-            case ast.Attribute():
-                match e.value:
-                    case ast.Attribute():
-                        val = self._parse_foreign_attribute(e.value)
-                        return ForeignAttribute(val.name, [NamedId(e.attr)] + val.attrs, loc)
-                    case ast.Name():
-                        name = self._parse_id(e.value)
-                        if isinstance(name, UnderscoreId):
-                            raise FPyParserError(loc, 'FPy foreign attribute must begin with a named identifier', e)
-                        return ForeignAttribute(name, [NamedId(e.attr)], loc)
-                    case _:
-                        raise FPyParserError(loc, 'FPy foreign attribute must begin with an identifier', e)
-            case _:
-                raise FPyParserError(loc, 'Not a valid foreign attribute', e)
+        value = self._parse_expr(e.value)
+        return Attribute(value, e.attr, loc)
 
-    def _parse_expr(self, e: ast.expr, *, allow_foreign: bool = False) -> Expr:
+    def _parse_expr(self, e: ast.expr) -> Expr:
         """Parse a Python expression."""
         loc = self._parse_location(e)
         match e:
             case ast.Name():
-                if e.id in _constants:
-                    return Constant(e.id, loc)
-                else:
-                    ident = self._parse_id(e)
-                    return Var(ident, loc)
+                ident = self._parse_id(e)
+                return Var(ident, loc)
             case ast.Constant():
                 return self._parse_constant(e)
             case ast.BoolOp():
@@ -435,47 +597,13 @@ class Parser:
             case ast.Compare():
                 return self._parse_compare(e)
             case ast.Call():
-                name = self._parse_call(e)
-                if name in _unary_table:
-                    if len(e.args) != 1:
-                        raise FPyParserError(loc, 'FPy unary operator expects one argument', e)
-                    arg = self._parse_expr(e.args[0])
-                    cls1 = _unary_table[name]
-                    return cls1(arg, loc)
-                elif name in _binary_table:
-                    if len(e.args) != 2:
-                        raise FPyParserError(loc, 'FPy binary operator expects two arguments', e)
-                    lhs = self._parse_expr(e.args[0])
-                    rhs = self._parse_expr(e.args[1])
-                    cls2 = _binary_table[name]
-                    return cls2(lhs, rhs, loc)
-                elif name in _ternary_table:
-                    if len(e.args) != 3:
-                        raise FPyParserError(loc, 'FPy ternary operator expects three arguments', e)
-                    arg0 = self._parse_expr(e.args[0])
-                    arg1 = self._parse_expr(e.args[1])
-                    arg2 = self._parse_expr(e.args[2])
-                    cls3 = _ternary_table[name]
-                    return cls3(arg0, arg1, arg2, loc)
-                elif name in _nary_table:
-                    args = [self._parse_expr(arg) for arg in e.args]
-                    clsN = _nary_table[name]
-                    return clsN(args, loc)
-                elif name == 'len':
-                    arg = self._parse_expr(e.args[0])
-                    return Size(arg, Integer(0, loc), loc)
-                elif name == 'rational':
-                    return self._parse_rational(e)
-                elif name == 'hexfloat':
-                    return self._parse_hexfloat(e)
-                elif name == 'digits':
-                    return self._parse_digits(e)
-                else:
-                    return Call(name, [self._parse_expr(arg) for arg in e.args], loc)
+                return self._parse_call(e)
+            case ast.Attribute():
+                return self._parse_attribute(e)
             case ast.Tuple():
                 return TupleExpr([self._parse_expr(e) for e in e.elts], loc)
             case ast.List():
-                return TupleExpr([self._parse_expr(e) for e in e.elts], loc)
+                return ListExpr([self._parse_expr(e) for e in e.elts], loc)
             case ast.ListComp():
                 targets: list[Id | TupleBinding] = []
                 iterables: list[Expr] = []
@@ -484,19 +612,16 @@ class Parser:
                     targets.append(target)
                     iterables.append(iterable)
                 elt = self._parse_expr(e.elt)
-                return CompExpr(targets, iterables, elt, loc)
+                return ListComp(targets, iterables, elt, loc)
             case ast.Subscript():
-                value, slices = self._parse_subscript(e)
-                return TupleRef(value, slices, loc)
+                return self._parse_subscript(e)
             case ast.IfExp():
                 cond = self._parse_expr(e.test)
                 ift = self._parse_expr(e.body)
                 iff = self._parse_expr(e.orelse)
                 return IfExpr(cond, ift, iff, loc)
             case _:
-                if not allow_foreign:
-                    raise FPyParserError(loc, 'expression is unsupported in FPy', e)
-                return self._parse_foreign_attribute(e)
+                raise FPyParserError(loc, 'expression is unsupported in FPy', e)
 
     def _parse_tuple_target(self, target: ast.expr, e: ast.AST):
         loc = self._parse_location(target)
@@ -539,49 +664,11 @@ class Parser:
             case None:
                 return UnderscoreId()
             case ast.Name():
-                return var.id
+                return NamedId(var.id)
             case _:
                 loc = self._parse_location(var)
                 raise FPyParserError(loc, '`Context` can only be optionally bound to an identifier`', var, item)
 
-    def _parse_contextexpr(self, item: ast.withitem):
-        e = item.context_expr
-        loc = self._parse_location(e)
-        match e:
-            case ast.Name():
-                name = self._parse_id(e)
-                return Var(name, loc)
-            case ast.Call():
-                # parse constructor
-                match e.func:
-                    case ast.Name():
-                        name = self._parse_id(e.func)
-                        if isinstance(name, UnderscoreId):
-                            raise FPyParserError(loc, 'Context constructor must be a named identifier or foreign attribute', e)
-                        ctor = Var(name, loc)
-                    case ast.Attribute():
-                        ctor = self._parse_foreign_attribute(e.func)
-                    case _:
-                        raise FPyParserError(loc, 'Context constructor must be a named identifier or foreign attribute', e)
-                # parse positional arguments
-                pos_args: list[Expr] = []
-                for arg in e.args:
-                    match arg:
-                        case ast.Starred():
-                            raise FPyParserError(loc, 'FPy does not support starred arguments', arg, e)
-                        case _:
-                            pos_args.append(self._parse_expr(arg, allow_foreign=True))
-                # parse keyword arguments
-                kw_args: list[tuple[str, Expr]] = []
-                for kwd in e.keywords:
-                    if kwd.arg is None:
-                        raise FPyParserError(loc, 'FPy does not support unnamed keyword arguments', kwd, e)
-                    v = self._parse_expr(kwd.value, allow_foreign=True)
-                    kw_args.append((kwd.arg, v))
-                return ContextExpr(ctor, pos_args, kw_args, loc)
-            case _:
-                raise FPyParserError(loc, 'FPy expects a valid context expression', e, item)
-
     def _parse_augassign(self, stmt: ast.AugAssign):
         loc = self._parse_location(stmt)
         if not isinstance(stmt.target, ast.Name):
@@ -593,20 +680,20 @@ class Parser:
 
         match stmt.op:
             case ast.Add():
-                cls: type[BinaryOp] = Add
+                value = self._parse_expr(stmt.value)
+                e: Expr = Add(Var(ident, loc), value, loc)
             case ast.Sub():
-                cls = Sub
+                value = self._parse_expr(stmt.value)
+                e = Sub(Var(ident, loc), value, loc)
             case ast.Mult():
-                cls = Mul
+                value = self._parse_expr(stmt.value)
+                e = Mul(Var(ident, loc), value, loc)
             case ast.Div():
-                cls = Div
-            case ast.Mod():
-                cls = Fmod
+                value = self._parse_expr(stmt.value)
+                e = Div(Var(ident, loc), value, loc)
             case _:
                 raise FPyParserError(loc, 'Unsupported operator-assignment in FPy', stmt)
 
-        value = self._parse_expr(stmt.value)
-        e = cls(Var(ident, loc), value, loc)
         return Assign(ident, None, e, loc)
 
     def _parse_statement(self, stmt: ast.stmt) -> Stmt:
@@ -641,7 +728,7 @@ class Parser:
                         value = self._parse_expr(stmt.value)
                         return Assign(binding, None, value, loc)
                     case ast.Subscript():
-                        var, slices = self._parse_subscript(target)
+                        var, slices = self._parse_subscript_target(target)
                         if not isinstance(var, Var):
                             raise FPyParserError(loc, 'FPy expects a variable', target, stmt)
                         value = self._parse_expr(stmt.value)
@@ -679,17 +766,21 @@ class Parser:
                     raise FPyParserError(loc, 'FPy only supports with statements with a single item', stmt)
                 item = stmt.items[0]
                 name = self._parse_contextname(item)
-                ctx = self._parse_contextexpr(item)
+                ctx = self._parse_expr(item.context_expr)
                 block = self._parse_statements(stmt.body)
                 return ContextStmt(name, ctx, block, loc)
             case ast.Assert():
                 test = self._parse_expr(stmt.test)
-                if stmt.msg is not None:
-                    raise FPyParserError(loc, 'FPy does not support assert messages', stmt)
-                return AssertStmt(test, None, loc)
+                if stmt.msg is None:
+                    return AssertStmt(test, None, loc)
+                else:
+                    msg = self._parse_expr(stmt.msg)
+                    return AssertStmt(test, msg, loc)
             case ast.Expr():
                 e = self._parse_expr(stmt.value)
                 return EffectStmt(e, loc)
+            case ast.Pass():
+                return PassStmt(loc)
             case _:
                 raise FPyParserError(loc, 'statement is unsupported in FPy', stmt)
 
@@ -714,22 +805,25 @@ class Parser:
 
         return args
 
+    def _parse_returns(self, e: ast.expr):
+        return self._parse_type_annotation(e)
+
     def _parse_lambda(self, f: ast.Lambda):
         """Parse a Python lambda expression."""
         loc = self._parse_location(f)
         args = self._parse_arguments(f.args.args)
         expr = self._parse_expr(f.body)
         block = StmtBlock([ReturnStmt(expr, expr.loc)])
-        return FuncDef('pre', args, block, loc=loc)
+        return FuncDef('pre', args, set(), None, block, None, {}, ForeignEnv.default(), loc=loc)
 
-    def _parse_function(self, f: ast.FunctionDef):
+    def _parse_function(self, f: ast.FunctionDef, env: ForeignEnv):
         """Parse a Python function definition."""
         loc = self._parse_location(f)
 
         # check arguments are only positional
         pos_args = f.args.posonlyargs + f.args.args
         if f.args.vararg:
-            raise FPyParserError(loc, 'FPy does not support variary arguments', f, f.args.vararg)
+            raise FPyParserError(loc, 'FPy does not support variadic arguments', f, f.args.vararg)
         if f.args.kwarg:
             raise FPyParserError(loc, 'FPy does not support keyword arguments', f, f.args.kwarg)
 
@@ -745,24 +839,71 @@ class Parser:
         block = self._parse_statements(body)
 
         # return AST and decorator list
-        func = FuncDef(f.name, args, block, loc=loc)
+        func = FuncDef(f.name, args, set(), None, block, None, {}, env, loc=loc)
         return func, f.decorator_list
 
     def _eval(
         self,
         e: ast.expr,
-        globals: Optional[Mapping[str, Any]] = None,
-        locals: Optional[Mapping[str, object]] = None
+        globals: Mapping[str, Any] | None = None,
+        locals: Mapping[str, object] | None = None
     ):
         globals = None if globals is None else dict(globals)
         return eval(ast.unparse(e), globals, locals)
 
+    def _start_parse(self):
+        start_loc = Location(self.name, self.start_line, 0, self.start_line, 0)
+
+        mod = ast.parse(self.source, self.name)
+        if len(mod.body) > 1:
+            raise FPyParserError(start_loc, 'FPy only supports single function definitions', mod)
+
+        ptree = mod.body[0]
+        if not isinstance(ptree, ast.FunctionDef):
+            raise FPyParserError(start_loc, 'FPy only supports single function definitions', mod)
+
+        return ptree
+
+
+    def parse_function(self, env: ForeignEnv):
+        """Parses `self.source` as an FPy `FunctionDef`."""
+        ptree = self._start_parse()
+        return self._parse_function(ptree, env)
+
+    def parse_signature(self, ignore_ctx: bool = False):
+        """Parses `self.source` to extract the arguments."""
+        f = self._start_parse()
+        loc = self._parse_location(f)
+
+        # check arguments are only positional
+        pos_args = f.args.posonlyargs + f.args.args
+        if f.args.vararg:
+            raise FPyParserError(loc, 'FPy does not support variadic arguments', f, f.args.vararg)
+        if f.args.kwarg:
+            raise FPyParserError(loc, 'FPy does not support keyword arguments', f, f.args.kwarg)
+
+        # check that there's a return annotation
+        if f.returns is None:
+            raise FPyParserError(loc, 'FPy requires a return annotation', f, f.returns)
+
+        # prune context argument
+        if ignore_ctx and len(pos_args) >= 1:
+            if pos_args[-1].arg == 'ctx':
+                pos_args = pos_args[:-1]
+
+        # parse arguments and returns
+        args = self._parse_arguments(pos_args)
+        returns = self._parse_returns(f.returns)
+
+        arg_types = [arg.type for arg in args]
+        return arg_types, returns
+
     def find_decorator(
         self,
         decorator_list: list[ast.expr],
         decorator: Any,
-        globals: Optional[Mapping[str, Any]] = None,
-        locals: Optional[Mapping[str, object]] = None
+        globals: Mapping[str, Any] | None = None,
+        locals: Mapping[str, object] | None = None
     ):
         """Returns the decorator AST for a particular decorator"""
         for dec in reversed(decorator_list):
@@ -771,7 +912,7 @@ class Parser:
                     f = self._eval(dec.func, globals=globals, locals=locals)
                     if isinstance(f, FunctionType) and f == decorator:
                         return dec
-                case ast.Name():
+                case ast.Name() | ast.Attribute():
                     f = self._eval(dec, globals=globals, locals=locals)
                     if isinstance(f, FunctionType) and f == decorator:
                         return dec
@@ -789,7 +930,7 @@ class Parser:
         - `pre`: a precondition expression
         """
         match decorator:
-            case ast.Name():
+            case ast.Name() | ast.Attribute():
                 return {}
             case ast.Call():
                 if decorator.args != []:
