@@ -6,13 +6,7 @@ from fractions import Fraction
 
 from ..ast import *
 from ..function import Function
-from ..utils import (
-    default_repr,
-    decnum_to_fraction,
-    digits_to_fraction,
-    hexnum_to_fraction,
-    sliding_window
-)
+from ..utils import default_repr, sliding_window
 
 from .pattern import Pattern, ExprPattern, StmtPattern
 from .subst import Subst
@@ -136,12 +130,12 @@ class _MatcherInst(Visitor):
 
     def _visit_decnum(self, e: Decnum, pat: Decnum):
         # this is a semantic match, not a syntactic match!
-        if decnum_to_fraction(e.val) != decnum_to_fraction(pat.val):
+        if e.as_rational() != pat.as_rational():
             raise _MatchFailure(f'matching {pat} against {e}')
 
     def _visit_hexnum(self, e: Hexnum, pat: Hexnum):
         # this is a semantic match, not a syntactic match!
-        if hexnum_to_fraction(e.val) != hexnum_to_fraction(pat.val):
+        if e.as_rational() != pat.as_rational():
             raise _MatchFailure(f'matching {pat} against {e}')
 
     def _visit_integer(self, e: Integer, pat: Integer):
@@ -150,18 +144,16 @@ class _MatcherInst(Visitor):
 
     def _visit_rational(self, e: Rational, pat: Rational):
         # this is a semantic match, not a syntactic match!
-        if Fraction(e.p, e.q) != Fraction(pat.p, pat.q):
-            raise _MatchFailure(f'matching {pat} against {e}')
-
-    def _visit_constant(self, e: Constant, pat: Constant):
-        if e.val != pat.val:
+        if e.as_rational() != pat.as_rational():
             raise _MatchFailure(f'matching {pat} against {e}')
 
     def _visit_digits(self, e: Digits, pat: Digits):
         # this is a semantic match, not a syntactic match!
-        f1 = digits_to_fraction(e.m, e.e, e.b)
-        f2 = digits_to_fraction(e.m, e.e, e.b)
-        if f1 != f2:
+        if e.as_rational() != pat.as_rational():
+            raise _MatchFailure(f'matching {pat} against {e}')
+
+    def _visit_nullaryop(self, e: NullaryOp, pat: NullaryOp):
+        if type(e) is not type(pat):
             raise _MatchFailure(f'matching {pat} against {e}')
 
     def _visit_unaryop(self, e: UnaryOp, pat: UnaryOp):
@@ -189,11 +181,29 @@ class _MatcherInst(Visitor):
             self._visit_expr(c1, c2)
 
     def _visit_call(self, e: Call, pat: Call):
-        if e.name != pat.name or len(e.args) != len(pat.args):
+        # check shape of call
+        if len(e.args) != len(pat.args) or len(e.kwargs) != len(pat.kwargs):
             raise _MatchFailure(f'matching {pat} against {e}')
+
+        # check function symbol
+        match e.fn, pat.fn:
+            case None, None:
+                if e.func != pat.func:
+                    raise _MatchFailure(f'matching {pat} against {e}')
+            case _, _:
+                if e.fn != pat.fn:
+                    raise _MatchFailure(f'matching {pat} against {e}')
+
+        # check arguments
         for c1, c2 in zip(e.args, pat.args):
             self._visit_expr(c1, c2)
 
+        # check keyword arguments
+        for (k1, v1), (k2, v2) in zip(e.kwargs, pat.kwargs):
+            if k1 != k2:
+                raise _MatchFailure(f'matching {pat} against {e}')
+            self._visit_expr(v1, v2)
+
     def _visit_compare(self, e: Compare, pat: Compare):
         # TODO: is matching on a subset of operations valid?
         # check if the operation sequence is the same
@@ -205,28 +215,50 @@ class _MatcherInst(Visitor):
 
     def _visit_tuple_expr(self, e: TupleExpr, pat: TupleExpr):
         # check if #elements are the same
-        if len(e.args) != len(pat.args):
+        if len(e.elts) != len(pat.elts):
             raise _MatchFailure(f'matching {pat} against {e}')
         # check if elements are the same
-        for c1, c2 in zip(e.args, pat.args):
+        for c1, c2 in zip(e.elts, pat.elts):
             self._visit_expr(c1, c2)
 
-    def _visit_tuple_ref(self, e: TupleRef, pat: TupleRef):
-        if len(e.slices) != len(pat.slices):
+    def _visit_list_expr(self, e: ListExpr, pat: ListExpr):
+        # check if #elements are the same
+        if len(e.elts) != len(pat.elts):
+            raise _MatchFailure(f'matching {pat} against {e}')
+        # check if elements are the same
+        for c1, c2 in zip(e.elts, pat.elts):
+            self._visit_expr(c1, c2)
+
+    def _visit_list_ref(self, e: ListRef, pat: ListRef):
+        self._visit_expr(e.value, pat.value)
+        self._visit_expr(e.index, pat.index)
+
+    def _visit_list_slice(self, e: ListSlice, pat: ListSlice):
+        self._visit_expr(e.value, pat.value)
+        match e.start, pat.start:
+            case None, None:
+                pass
+            case Expr(), Expr():
+                self._visit_expr(e.start, pat.start)
+            case _:
+                raise _MatchFailure(f'matching {pat} against {e}')
+        match e.stop, pat.stop:
+            case None, None:
+                pass
+            case Expr(), Expr():
+                self._visit_expr(e.stop, pat.stop)
+            case _:
+                raise _MatchFailure(f'matching {pat} against {e}')
+
+    def _visit_list_set(self, e: ListSet, pat: ListSet):
+        if len(e.indices) != len(pat.indices):
             raise _MatchFailure(f'matching {pat} against {e}')
         self._visit_expr(e.value, pat.value)
-        for s1, s2 in zip(e.slices, pat.slices):
+        for s1, s2 in zip(e.indices, pat.indices):
             self._visit_expr(s1, s2)
+        self._visit_expr(e.expr, pat.expr)
 
-    def _visit_tuple_set(self, e: TupleSet, pat: TupleSet):
-        if len(e.slices) != len(pat.slices):
-            raise _MatchFailure(f'matching {pat} against {e}')
-        self._visit_expr(e.array, pat.array)
-        for s1, s2 in zip(e.slices, pat.slices):
-            self._visit_expr(s1, s2)
-        self._visit_expr(e.value, pat.value)
-
-    def _visit_comp_expr(self, e: CompExpr, pat: CompExpr):
+    def _visit_list_comp(self, e: ListComp, pat: ListComp):
         if len(e.iterables) != len(pat.iterables):
             raise _MatchFailure(f'matching {pat} against {e}')
         for target, ptarget in zip(e.targets, pat.targets):
@@ -246,21 +278,10 @@ class _MatcherInst(Visitor):
         self._visit_expr(e.ift, pat.ift)
         self._visit_expr(e.iff, pat.iff)
 
-    def _visit_context_expr(self, e: ContextExpr, ctx: ContextExpr):
-        if len(e.args) != len(ctx.args):
-            raise _MatchFailure(f'matching {ctx} against {e}')
-        for c1, c2 in zip(e.args, ctx.args):
-            match c1, c2:
-                case ForeignAttribute(), ForeignAttribute():
-                    if c1 != c2:
-                        raise _MatchFailure(f'matching {ctx} against {e}')
-                case Expr(), Expr():
-                    # check if args are the same
-                    self._visit_expr(c1, c2)
-                case (ForeignAttribute(), _) | (_, ForeignAttribute()):
-                    raise _MatchFailure(f'matching {ctx} against {e}')
-                case _, _:
-                    raise RuntimeError(f'unreachable case: {c1} vs {c2}')
+    def _visit_attribute(self, e: Attribute, pat: Attribute):
+        if e.attr != pat.attr:
+            raise _MatchFailure(f'matching {pat} against {e}')
+        self._visit_expr(e.value, pat.value)
 
     def _visit_binding(self, binding: Id | TupleBinding, pat: Id | TupleBinding):
         match binding, pat:
@@ -272,7 +293,7 @@ class _MatcherInst(Visitor):
                 raise _MatchFailure(f'matching {pat} against {binding}')
 
     def _visit_assign(self, stmt: Assign, pat: Assign):
-        self._visit_binding(stmt.binding, pat.binding)
+        self._visit_binding(stmt.target, pat.target)
         self._visit_expr(stmt.expr, pat.expr)
 
     def _visit_tuple_binding(self, binding: TupleBinding, pat: TupleBinding):
@@ -294,7 +315,7 @@ class _MatcherInst(Visitor):
 
     def _visit_indexed_assign(self, stmt: IndexedAssign, pat: IndexedAssign):
         self._visit_target(stmt.var, pat.var)
-        for e, p in zip(stmt.slices, pat.slices):
+        for e, p in zip(stmt.indices, pat.indices):
             self._visit_expr(e, p)
         self._visit_expr(stmt.expr, pat.expr)
 
@@ -321,6 +342,7 @@ class _MatcherInst(Visitor):
         self._visit_block(stmt.body, pat.body)
 
     def _visit_assert(self, stmt: AssertStmt, pat: AssertStmt):
+        # TODO: message?
         self._visit_expr(stmt.test, pat.test)
 
     def _visit_effect(self, stmt: EffectStmt, pat: EffectStmt):
@@ -329,6 +351,9 @@ class _MatcherInst(Visitor):
     def _visit_return(self, stmt: ReturnStmt, pat: ReturnStmt):
         self._visit_expr(stmt.expr, pat.expr)
 
+    def _visit_pass(self, stmt: PassStmt, pat: ReturnStmt):
+        pass
+
     def _visit_block(self, block: StmtBlock, pat: StmtBlock):
         # check length of block
         if len(block.stmts) != len(pat.stmts):
@@ -406,7 +431,7 @@ class _StmtMatcherEngine(DefaultVisitor):
         pattern_block = self.pattern.block
         for i, stmts in enumerate(sliding_window(block.stmts, len(pattern_block.stmts))):
             # check if the pattern matches
-            m = _MatcherInst(self.pattern, StmtBlock(stmts))
+            m = _MatcherInst(self.pattern, StmtBlock(list(stmts)))
             pmatch = m.match()
             if pmatch is not None:
                 if not isinstance(pmatch, StmtMatch):
