@@ -1,9 +1,9 @@
 """Common utilities for the FPy infrastructure."""
 
-from .bits import bitmask, float_to_bits, bits_to_float
+from .bits import bitmask, is_power_of_two, float_to_bits, bits_to_float
 from .compare import CompareOp
 from .decorator import default_repr, rcomparable, enum_repr
-from .error import FPySyntaxError, raise_type_error
+from .default import DefaultOr, DEFAULT
 
 from .float_params import (
     FP64_NBITS,
@@ -15,9 +15,14 @@ from .float_params import (
     FP64_EXPMAX,
     FP64_EXPMIN,
     FP64_BIAS,
+    FP64_SMASK,
+    FP64_EMASK,
+    FP64_MMASK,
+    FP64_EONES,
+    FP64_IMPLICIT1
 )
 
-from .fractions import fraction, digits_to_fraction, decnum_to_fraction, hexnum_to_fraction
+from .fractions import digits_to_fraction, decnum_to_fraction, hexnum_to_fraction, is_dyadic
 from .gensym import Gensym
 from .identifier import Id, NamedId, UnderscoreId, SourceId
 from .inspect import has_keyword
@@ -25,3 +30,5 @@ from .iterator import sliding_window
 from .location import Location
 from .ordering import Ordering
 from .string import pythonize_id
+from .uninit import UNINIT
+from .unionfind import Unionfind
